//
//  Copyright 2013 Canonical Ltd.
//
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1AssetManager.h"

#import "U1Logging.h"

#import "U1FolderNode.h"
#import "U1FileNode.h"

#define BATCHSIZE 2000

@interface U1AssetManager ()
- (BOOL)findOrCreateU1AssetsForAssets:(NSArray*)batch
							  inGroup:(ALAssetsGroup *)group
						  withContext:(NSManagedObjectContext *)context
						   assetBlock:(void(^)(U1Asset *u1asset, BOOL isNew))assetBlock;

@end

@implementation U1AssetManager

+ (U1AssetManager *)sharedAssetManager{
	static U1AssetManager *sharedAssetManager = nil;
	if (sharedAssetManager == nil)
	{
		sharedAssetManager = [[U1AssetManager alloc] init];
	}
	return sharedAssetManager;
}

- (id)init{
    if (!(self = [super init])){
		return nil;
	}
	self.dataRepository = [U1DataRepository sharedDataRepository];
	self.assetsLibrary = [[[ALAssetsLibrary alloc] init] autorelease];
	return self;
}

- (void)updatePhotoAssetsWithAsyncBlock:(void(^)(U1Asset *u1asset, BOOL isNew))assetBlock groupCompletionBlock:(void(^)(BOOL success))groupCompletionBlock{
	
	NSArray *urlSortDescriptors = [NSArray arrayWithObject:[NSSortDescriptor sortDescriptorWithKey:@"defaultRepresentation.url.absoluteString" ascending:YES]];

	[self.assetsLibrary enumerateGroupsWithTypes:ALAssetsGroupSavedPhotos
		usingBlock:^(ALAssetsGroup *group, BOOL *stop) {

			// guarantee to do work on background queue. Although enumerateGroupsWithTypes is asynchronous due to asking for
			// permission, it apparently will execute the block synchronously if it already has permission.
			[self.dataRepository dispatchAsyncBlockWithManagedObjectContext:^(NSManagedObjectContext *localContext) {
				
				__block BOOL success = YES;
				
				NSMutableArray *batch = [NSMutableArray arrayWithCapacity:BATCHSIZE];
				
				[group enumerateAssetsUsingBlock:^(ALAsset *asset, NSUInteger index, BOOL *stop) {
					
					if (asset == nil){
						return;
					}
					
					[batch addObject:asset];
					
					if ([batch count] < BATCHSIZE) return;
					
					[batch sortUsingDescriptors:urlSortDescriptors];
					success = [self findOrCreateU1AssetsForAssets:batch inGroup:group withContext:localContext assetBlock:assetBlock];
					[batch removeAllObjects];
					
					if(!success){
						*stop = YES;
					}
					
				}];
				
				if(!success){
					*stop = YES;
					return;
				}
				
				if ([batch count] > 0){
					[batch sortUsingDescriptors:urlSortDescriptors];
					success = [self findOrCreateU1AssetsForAssets:batch inGroup:group withContext:localContext assetBlock:assetBlock];
				}
				
				// save context so that any items created (e.g. using findOrCreateU1AssetsForAssets:...) by an assetBlock are saved in main context. This way the groupCompletionBlock can safely access the new objects with a different context.
				NSError *error;
				if(![localContext save:&error]){
					CLS_LOG(@"Error saving context before group completion block is called: %@", error);
					*stop = YES;
					return;
				}
				
				groupCompletionBlock(success);
			}];
										  
	} failureBlock:^(NSError *error) {
		NSLog(@"Error walking through photo groups: %@", error);
	}];
}


- (BOOL)findOrCreateU1AssetsForAssets:(NSArray*)batch
							  inGroup:(ALAssetsGroup *)group
						  withContext:(NSManagedObjectContext *)context
						   assetBlock:(void(^)(U1Asset *u1asset, BOOL isNew))assetBlock{

	NSMutableArray *URLs = [batch valueForKeyPath:@"defaultRepresentation.url.absoluteString"];

	NSPredicate *urlInBatchPredicate = [NSPredicate predicateWithFormat:@"url in %@", URLs];
	NSError *error = nil;
	NSSortDescriptor *urlSortDescriptor = [NSSortDescriptor sortDescriptorWithKey:@"url" ascending:YES];
	NSArray *existingU1Assets = [self.dataRepository resultsForEntityClass:[U1Asset class]
																 inContext:context
															 withPredicate:urlInBatchPredicate
														   sortDescriptors:[NSArray arrayWithObject:urlSortDescriptor]
																	 error:&error];
	if(existingU1Assets == nil){
		CLS_LOG(@"Error looking for existing U1Assets: %@", error);
		return NO;
	}
	
	NSEnumerator *existing_U1AssetEnumerator = [existingU1Assets objectEnumerator];
	__block U1Asset *cur_U1Asset = [existing_U1AssetEnumerator nextObject];
	
	__block NSMutableArray *oldAssetsToUpload = [NSMutableArray arrayWithCapacity:[batch count]];
	__block NSMutableArray *newAssetsToUpload = [NSMutableArray arrayWithCapacity:[batch count]];
	
	[batch enumerateObjectsUsingBlock:^(ALAsset *asset, NSUInteger idx, BOOL *stop) {
		
		NSString *asset_url = [[[asset defaultRepresentation] url] absoluteString];
		
		
		// advance cursor past previous matching u1asset
		while (cur_U1Asset && [cur_U1Asset.url compare:asset_url] == NSOrderedAscending){
			cur_U1Asset = [existing_U1AssetEnumerator nextObject];
		}
		
		if ([cur_U1Asset.url isEqualToString:asset_url]){
			[oldAssetsToUpload addObject:cur_U1Asset];
			
		}else{
			
			NSString *assetType = [asset valueForProperty:ALAssetPropertyType];
			if([assetType isEqualToString:ALAssetTypePhoto]){
				
				U1Asset *newU1Asset = [self createU1AssetWithAsset:asset
															 group:group
														 URLString:asset_url
														  fileName:[[asset defaultRepresentation] filename]
														   context:context];
				[newAssetsToUpload addObject:newU1Asset];
			}
		}
		
	}];
	
	if (![context save:&error]){
		CLS_LOG(@"failed to save batch of u1assets: %@", error);
		return NO;
	}
	
	// run assetBlock on all assets to upload - only after successful save that commits changes to main context.
	[oldAssetsToUpload enumerateObjectsUsingBlock:^(U1Asset *u1asset, NSUInteger idx, BOOL *stop) {
		assetBlock(u1asset, NO);
	}];
	[newAssetsToUpload enumerateObjectsUsingBlock:^(U1Asset *u1asset, NSUInteger idx, BOOL *stop) {
		assetBlock(u1asset, YES);
	}];
	
	
	return YES;
}


- (U1Asset *)createU1AssetWithAsset:(ALAsset *)asset group:(ALAssetsGroup *)group URLString:(NSString *)urlString fileName:(NSString *)filename  context:(NSManagedObjectContext*)context{
	
	U1Asset *newAsset = [U1Asset insertInManagedObjectContext:context];
	newAsset.groupId = [group valueForProperty: ALAssetsGroupPropertyPersistentID];
	newAsset.url = urlString;
	newAsset.filename = filename;
	
	return newAsset;
}


- (void)findOrCreateMatchingFileNodeForU1Asset:(U1Asset *)u1asset inFolderNode:(U1FolderNode *)folderNode context:(NSManagedObjectContext*)context{
	
	NSString *resourcePath = [folderNode.resourcePath stringByAppendingPathComponent:u1asset.filename];
	NSString *contentPath = [folderNode.contentPath stringByAppendingPathComponent:u1asset.filename];
	
	__block BOOL found = NO;
	__block U1FileNode *node = nil;
	
	[folderNode.children enumerateObjectsUsingBlock:^(U1FileNode *childNode, BOOL *stop) {
		if ([childNode.name isEqualToString:u1asset.filename]) {
			found = YES;
			u1asset.generation = childNode.generation;
			u1asset.fileNode = childNode;
			*stop = YES;
		}
	}];
	if (!found){
		node = [U1FileNode insertInManagedObjectContext:context];
		node.resourcePath = resourcePath;
		node.kind = @"file";
		node.parent = folderNode;
		node.contentPath = contentPath;
		node.asset = u1asset;
	}
}


@end
