//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import <AssetsLibrary/AssetsLibrary.h>
#import <MobileCoreServices/UTType.h>

#import "U1Logging.h"

#import "Reachability.h"

#import "U1AccountManager.h"
#import "U1Asset.h"
#import "U1AssetManager.h"
#import "U1DataRepository.h"
#import "U1AutoUploadsManager.h"
#import "U1FileNode.h"
#import "U1FilesClient.h"
#import "U1FolderNode.h"
#import "U1Volume.h"


@interface U1AutoUploadsManager ()
@property (retain) U1FilesClient *filesClient;
@property (retain) NSOperationQueue *autoUploadPreparationQueue;
@property (retain) NSManagedObjectID *autoUploadsFolderID;
- (void)scanAssetLibraryForUploads;
- (void)enqueueUploadOperationsForAssetIDs:(NSArray *)assetIDsToUpload;
- (BOOL)isPending:(NSString *)filename;
- (NSString*)remoteUploadFolderPath;
- (void)fetchRemoteUploadFolderContents;
- (void)queuePreviousUploads;
- (void)volumesFetched:(NSNotification *)notification;
@end


@implementation U1AutoUploadsManager

@synthesize filesClient, dataRepository, assetsLibrary;
@synthesize autoUploadPreparationQueue, autoUploadsFolderID;
@synthesize remoteUploadFolder;

+ (U1AutoUploadsManager *)sharedAutoUploadsManager;
{
	static U1AutoUploadsManager *sharedAutoUploadsManager = nil;
	if (sharedAutoUploadsManager == nil)
	{
		sharedAutoUploadsManager = [[U1AutoUploadsManager alloc] init];
	}
	return sharedAutoUploadsManager;
}

- (id)init
{
    if (!(self = [super init]))
		return nil;

	
	filesClient = [[U1FilesClient sharedFilesClient] retain];
	autoUploadPreparationQueue = [[NSOperationQueue alloc] init];
	[self.autoUploadPreparationQueue setMaxConcurrentOperationCount:1];
    assetsLibrary = [[ALAssetsLibrary alloc] init];
	// Work around to ensure that notifications actually fire
	[self.assetsLibrary writeImageToSavedPhotosAlbum:nil metadata:nil completionBlock:^(NSURL *assetURL, NSError *error) { }];
	dataRepository = [U1DataRepository sharedDataRepository];

	NSNotificationCenter *defaultCenter = [NSNotificationCenter defaultCenter];
	[defaultCenter addObserver:self selector:@selector(libraryChanged:) name:ALAssetsLibraryChangedNotification object:self.assetsLibrary];
	[defaultCenter addObserver:self selector:@selector(reachabilityChanged:) name:kReachabilityChangedNotification object:nil];
	[defaultCenter addObserver:self selector:@selector(volumesFetched:) name:@"volumesFetched" object:nil];
	[defaultCenter addObserver:self selector:@selector(didSignOut:) name:U1AccountManagerDidSignOut object:nil];

	NSUserDefaults *standardDefaults = [NSUserDefaults standardUserDefaults];
	[standardDefaults addObserver:self forKeyPath:@"auto_upload" options:NSKeyValueObservingOptionNew context:NULL];

    return self;
}

- (void)dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	[filesClient release];
	[assetsLibrary release];
	[autoUploadPreparationQueue release];
	[remoteUploadFolder release];
	[autoUploadsFolderID release];
    [super dealloc];
}

- (void)checkForNewAssets;
{
	[self scanAssetLibraryForUploads];
}

- (void)libraryChanged:(NSNotification*)notification;
{
	[self scanAssetLibraryForUploads];
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];

	if (object == defaults && [keyPath isEqualToString:@"auto_upload"])
	{
		if ([defaults boolForKey:@"auto_upload"])
		{
			// Turned it on, start your engines
			[self.autoUploadPreparationQueue setSuspended:NO];
			// trigger the fetch volumes with a no-op completionBlock. XXX: Minorly skeezy.
			[self.filesClient volumesWithResultsBlock:^(U1Volume *rootVolume, NSFetchedResultsController *cloudFoldersResultsController, U1DataSourceType resultsDataSourceType) {} errorBlock:^(NSError *error) {}];
		}
		else
		{
			[self.autoUploadPreparationQueue setSuspended:YES];
		}
	}
}

- (void)reachabilityChanged:(NSNotification*)notification;
{
	Reachability *reachability = [notification object];
	NetworkStatus status = [reachability currentReachabilityStatus];
	switch (status)
	{
		case ReachableViaWiFi:
			[self.autoUploadPreparationQueue setSuspended:NO];
			break;
		case ReachableViaWWAN:
		{
			BOOL shouldSuspend = [[NSUserDefaults standardUserDefaults] boolForKey:@"wifi_only"];
			[self.autoUploadPreparationQueue setSuspended:shouldSuspend];
			break;
		}
		case NotReachable:
			[self.autoUploadPreparationQueue setSuspended:YES];
			break;
	}
}

- (void)volumesFetched:(NSNotification *)notification;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];

	if ([defaults boolForKey:@"auto_upload"])
	{
		NSArray *volumes = [notification object];
		BOOL found = NO;
		U1Volume *volume = nil;
		for (U1Volume *testVolume in volumes) {
			if ([testVolume.nodePath isEqualToString:[self remoteUploadFolderPath]])
			{
				found = YES;
				volume = testVolume;
				break;
			}
		}
		if (!found)
		{
			static BOOL creatingUploadVolume = NO;
			
			if (creatingUploadVolume)
				return;
			
			creatingUploadVolume = YES;
			
			[self.filesClient createVolumeAtPath:[self remoteUploadFolderPath]
					  completionBlockOnMainQueue:^(U1Volume *volume, NSError *error) {
									 creatingUploadVolume = NO;
									 if (error == nil)
									 {
										 self.remoteUploadFolder = volume.rootFolder;
										 self.autoUploadsFolderID = [volume.rootFolder objectID];
										 [self fetchRemoteUploadFolderContents];
									 }
									 else
									 {
										 NSLog(@"Error ensuring upload folder: %@", error);
									 }
								 }];	
		}
		else
		{
			self.remoteUploadFolder = volume.rootFolder;
			self.autoUploadsFolderID = [volume.rootFolder objectID];
			[self fetchRemoteUploadFolderContents];
		}
	}
	// We received a list of volumes. If our upload folder is not among them, prime it.
}

- (void)didSignOut:(NSNotification*)notification;
{
	[self.autoUploadPreparationQueue cancelAllOperations];
}

- (void)fetchRemoteUploadFolderContents;
{
	// fetch the remote upload folder contents, and stash the children objects somewhere locally.
	[self.filesClient fetchFolderWithResourcePath:self.remoteUploadFolder.resourcePath
									 resultsBlock:^(U1Node *node, NSFetchedResultsController *childrenResultsController, U1DataSourceType resultsDataSourceType) {
										 if (resultsDataSourceType == U1DataSourceRemote){
											 // This block will be executed on the main queue. (if data source is local, it is executed on the calling queue)
											 NSError *err;
											 if(![childrenResultsController performFetch:&err]){
												 CLS_LOG(@"Error fetching contents of remote upload folder: %@", err);
											 }
											 
											 [self queuePreviousUploads];
											 [self checkForNewAssets];
										 }
									 }
									   errorBlock:^(NSError *error) {}];
}

- (void)scanAssetLibraryForUploads;
{
	static BOOL alreadyWalking = NO;
	
	if (alreadyWalking)
		return;
	
	alreadyWalking = YES;
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	
	if (![defaults boolForKey:@"auto_upload"]){
		return;
	}
	
	// save as IDs because assets are fetched/created on an async queue
	__block NSMutableArray *assetIDsToUpload = nil;

	[[U1AssetManager sharedAssetManager] updatePhotoAssetsWithAsyncBlock:^(U1Asset *u1asset, BOOL isNew){
	

		// this block is invoked on the same queue as the code that created the u1asset, so we use its context:
		NSManagedObjectContext *localContext = u1asset.managedObjectContext;
		U1FolderNode *autoUploadsFolder = (U1FolderNode*)[localContext objectWithID:self.autoUploadsFolderID];
		
		// ignore assets with fileNodes that are not in the autoUploadsFolder
		if (u1asset.fileNode != nil){
			if(u1asset.fileNode.parent == nil){
				CLS_LOG(@"WARNING: asset %@ has a fileNode with no parent. FileNode is %@", u1asset, u1asset.fileNode);
				return;
			}
			if(u1asset.fileNode.parent != autoUploadsFolder) {
				return;
			}
		}
		
		// do not upload existing u1assets with non-nil generations, they're already there.
		if(!isNew && u1asset.generation != nil) {
			return;
		}
		
		if (isNew) {
			[[U1AssetManager sharedAssetManager] findOrCreateMatchingFileNodeForU1Asset:u1asset inFolderNode:autoUploadsFolder context:localContext];
			
		}
		
		if(assetIDsToUpload == nil){
			assetIDsToUpload = [[NSMutableArray alloc] init];
		}

		[assetIDsToUpload addObject:[u1asset objectID]];
			
	} groupCompletionBlock:^(BOOL success){

		if (!success) {
			// U1Assets are not consistent, do not queue uploads to avoid later crashing.
			return;
		}
				
		if(assetIDsToUpload == nil){
			return;
		}
		
		[self enqueueUploadOperationsForAssetIDs:assetIDsToUpload];

		[assetIDsToUpload autorelease];
		assetIDsToUpload = nil;
		
		alreadyWalking = NO;
								
	}];
		
}


- (void)enqueueUploadOperationsForAssetIDs:(NSArray *)assetIDsToUpload{

	NSMutableArray *ops = [NSMutableArray arrayWithCapacity:[assetIDsToUpload count]];
	
	[[U1DataRepository sharedDataRepository] dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {

		[assetIDsToUpload enumerateObjectsUsingBlock:^(id u1assetID, NSUInteger idx, BOOL *stop) {
			U1Asset *asset = (U1Asset *)[context objectWithID:u1assetID];
			
			NSOperation *op = [self autoUploadOperationForAsset:asset
											  andRepresentation:nil
													  inContext:context];
			if (op){
				[ops addObject:op];
			}
		}];
	}];

	NSArray *sortedOps = [ops sortedArrayUsingDescriptors:[NSArray arrayWithObject:[NSSortDescriptor sortDescriptorWithKey:@"filename" ascending:YES]]];
	[sortedOps enumerateObjectsUsingBlock:^(U1AutoUploadOperation *operation, NSUInteger idx, BOOL *stop) {
		[self.autoUploadPreparationQueue addOperation:operation];
		
	}];
	
	
		
}

	
// This is run on the main queue
- (void)queuePreviousUploads;
{
	// We should poll our data repository for all U1Assets which have generation=nil. They may not all have U1FileNodes associated with them.
	// If the U1FileNode's parent equals our autoUploadsManager's remoteUploadFolder, then it's an auto-upload, and we handle it in checkForNewAssets.
	// Otherwise, kick it off immediately to the FilesClient and get it back in the queue.
	
	NSPredicate *p = [NSPredicate predicateWithFormat:@"generation = %@", nil];
	NSError *error = nil;
	NSSortDescriptor *sortBy = [NSSortDescriptor sortDescriptorWithKey:@"filename" ascending:YES];
	NSArray *previouslyQueuedU1Assets = [self.dataRepository resultsForEntityClass:[U1Asset class]
																 matchingPredicate:p
															   withSortDescriptors:[NSArray arrayWithObject:sortBy]
																			 error:&error];
	if(previouslyQueuedU1Assets == nil){
		CLS_LOG(@"Error fetching U1Asset records from main context: %@", error);
	}

	[previouslyQueuedU1Assets enumerateObjectsUsingBlock:^(U1Asset *u1asset, NSUInteger idx, BOOL *stop) {
		if(u1asset.fileNode == nil) {
			return;
		}
		if (u1asset.fileNode.parent == self.remoteUploadFolder)
		{
			// this case handled by checkForNewAssets
			return;
		}
		NSURL *imageURL = [NSURL URLWithString:u1asset.url];
		
		[self.assetsLibrary assetForURL:imageURL resultBlock:^(ALAsset *asset) {
			if (asset == nil)
			{
				NSLog(@"In the results block, but asset is nil!");
			}
			else
			{
				// examine u1asset on main queue, as it was fetched there:
				dispatch_async(dispatch_get_main_queue(), ^{
					
					[self.filesClient uploadAssetWithID:[u1asset objectID]
										   withPriority:NSOperationQueuePriorityVeryHigh
										  progressBlock:^(long long bytesUploaded, long long totalBytes) {}
							 completionBlockOnMainQueue:^(U1FileNode *updatedNode, NSError *error){
								 if(error != nil){
									 NSLog(@"error in upload for restarted auto-upload: %@", error);
								 }else{
									 
									 u1asset.generation = updatedNode.generation;
									 
									 NSError *err = NULL;
									 BOOL success = [dataRepository save:&err];
									 if(!success){
										 CLS_LOG(@"U1AutoUploadOperation: failed to save to repository: %@", err);
									 }
								 }
							 }];
				});
			}
		} failureBlock:^(NSError *error) {
			NSLog(@"Couldn't fetch asset from library: %@", error);
		}];
	}];
}

- (NSString *)remoteUploadFolderPath;
{
	return [NSString stringWithFormat:@"/~/Pictures - %@", [[UIDevice currentDevice] name]];
}


- (U1AutoUploadOperation *)autoUploadOperationForAsset:(U1Asset *)assetToUpload andRepresentation:(ALAssetRepresentation *)representation inContext:(NSManagedObjectContext *)context;
{
	// If the assetToUpload's generation is nil, but its fileNode's generation isn't, then the Node already exists on the server. Update the asset and move on
	if (assetToUpload.generation == nil && assetToUpload.fileNode.generation != nil)
	{
		assetToUpload.generation = assetToUpload.fileNode.generation;
		NSError *err;
		if(![context save:&err]){
			CLS_LOG(@"Error saving generation update to asset: %@",err);
		}
		return nil;
	}
	
	if([self isPending:assetToUpload.filename]){
		return nil;
	}
	
	if(assetToUpload.generation != nil){
		return nil;
	}
	
	NSLog(@"creating an autoupload op for asset ID %@", [assetToUpload objectID]);
	U1AutoUploadOperation *operation = [[U1AutoUploadOperation alloc] init];
	operation.assetID = [assetToUpload objectID];
	operation.folderID = [self.remoteUploadFolder objectID];
	operation.filename = assetToUpload.filename;
	operation.priority = NSOperationQueuePriorityNormal;
	return [operation autorelease];
	
}

- (int)numberOfAssets;
{
	NSError *error = nil;
	return [[self.dataRepository resultsForEntityClass:[U1Asset class]
									matchingPredicate:nil
								  withSortDescriptors:nil
												error:&error] count];
}

- (void)thumbnailForNode:(U1FileNode *)fileNode completionBlock:(void(^)(CGImageRef thumbnail))completionBlock;
{
	[self.assetsLibrary assetForURL:[NSURL URLWithString:fileNode.asset.url]
						resultBlock:^(ALAsset *asset) {
							completionBlock(asset.thumbnail);
						} failureBlock:^(NSError *error) {
							NSLog(@"Couldn't find the asset to get a thumbnail: %@", error);
						}];
}

- (BOOL)isPending:(NSString *)filename;
{
	__block BOOL isPending = NO;
	
	[[self.autoUploadPreparationQueue operations] enumerateObjectsUsingBlock:^(U1AutoUploadOperation *operation, NSUInteger idx, BOOL *stop) {
		if ([operation.filename isEqualToString:filename])
		{
			isPending = YES;
			*stop = YES;
		}
	}];
	
	return isPending;
}

- (int)pendingCount;
{
	__block int count = 0;
	
	[[self.autoUploadPreparationQueue operations] enumerateObjectsUsingBlock:^(U1AutoUploadOperation *operation, NSUInteger idx, BOOL *stop) {
		if (![operation isExecuting])
		{
			count++;
		}
	}];
	
	return count;
}

@end
