//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1DataRepository.h"

#import <CoreData/CoreData.h>

#import "U1Logging.h"

#import "U1Node.h"

static U1DataRepository *sharedDataRepository = nil;

@interface U1DataRepository ()


- (NSArray*)resultsForEntityClass:(Class)class inContext:(NSManagedObjectContext*)context withPredicate:(NSPredicate*)predicate sortDescriptors:(NSArray*)sortDescriptors error:(NSError**)error;

- (void)addPersistentStoreAtURL:(NSURL*)storeURL;

@end


@implementation U1DataRepository

@synthesize model, mainContext, storeCoordinator;

+ (U1DataRepository*)sharedDataRepository;
{
	if (sharedDataRepository == nil)
	{
		sharedDataRepository = [[self alloc] initWithStoreURL:[U1DataRepository databaseURLInDirectory:NSDocumentDirectory]];
	}
	return sharedDataRepository;
}

+ (NSURL *)databaseURLInDirectory:(NSUInteger)directory;
{
	NSURL *directoryURL = [[[NSFileManager defaultManager] URLsForDirectory:directory
																  inDomains:NSUserDomainMask] lastObject];
	return [NSURL URLWithString:@"U1Files.sqlite" relativeToURL:directoryURL];
}

- (void)addPersistentStoreAtURL:(NSURL*)storeURL;
{
	
	NSDictionary *options = [NSDictionary dictionaryWithObjectsAndKeys:
							 [NSNumber numberWithBool:YES], NSMigratePersistentStoresAutomaticallyOption,
							 [NSNumber numberWithBool:YES], NSInferMappingModelAutomaticallyOption,
							 nil];
	
	NSError *error = nil;
	if (![self.storeCoordinator addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:storeURL options:options error:&error])
	{
		CLS_LOG(@"Error opening persistent store: %@", error);
	}

}

- (id)initWithStoreURL:(NSURL*)storeURL;
{
	if (!(self = [super init]))
		return nil;
	
	NSURL *modelURL = [[NSBundle mainBundle] URLForResource:@"U1Files" withExtension:@"momd"];
	self.model = [[[NSManagedObjectModel alloc] initWithContentsOfURL:modelURL] autorelease];
	
	if (self.model == nil)
		return nil;
	
	self.storeCoordinator = [[[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:self.model] autorelease];
	if (!self.storeCoordinator){
		return nil;
	}

	[self addPersistentStoreAtURL:storeURL];

	self.mainContext = [[[NSManagedObjectContext alloc] init] autorelease];
	[self.mainContext setPersistentStoreCoordinator:self.storeCoordinator];
	[self.mainContext setMergePolicy:NSOverwriteMergePolicy];

	return self;
}

- (void)dealloc;
{
	[model release];
	[mainContext release];
	[storeCoordinator release];
	[super dealloc];
}

- (BOOL)save:(NSError**)error;
{
	return [self.mainContext save:error];
}

- (BOOL)reset:(NSError**)error;
{
	NSURL * storeURL = [U1DataRepository databaseURLInDirectory:NSDocumentDirectory];

	[self.mainContext lock];
	[self.mainContext reset];

	if ([self.storeCoordinator removePersistentStore:[[self.storeCoordinator persistentStores] lastObject] error:error])
	{
		if([[NSFileManager defaultManager] removeItemAtURL:storeURL error:error]){
			[self addPersistentStoreAtURL:storeURL];
		}else{
			CLS_LOG(@"problem removing file backing persistent store: %@", *error);
			[self.mainContext unlock];
			return NO;
		}
	}else{
		CLS_LOG(@"Problem removing persistent store in reset: %@", *error);
	}
	[self.mainContext unlock];
	return YES;
}

- (U1Node*)nodeWithResourcePath:(NSString *)resourcePath;
{
	NSParameterAssert(resourcePath);

	NSPredicate *p = [NSPredicate predicateWithFormat:@"resourcePath = %@", resourcePath];
	NSArray *nodes = [self resultsForEntityClass:[U1Node class] inContext:self.mainContext withPredicate:p sortDescriptors:nil error:NULL];
	U1Node *node = [nodes lastObject];
	return node;
}

- (NSArray*)resultsForEntityClass:(Class)entityClass matchingPredicate:(NSPredicate*)predicate withSortDescriptors:(NSArray*)sortDescriptors error:(NSError**)error;
{
	return [self resultsForEntityClass:entityClass inContext:self.mainContext withPredicate:predicate sortDescriptors:sortDescriptors error:error];
}

- (NSArray*)resultsForEntityClass:(Class)class inContext:(NSManagedObjectContext*)context withPredicate:(NSPredicate*)predicate sortDescriptors:(NSArray*)sortDescriptors error:(NSError**)error;
{
	NSParameterAssert(class != nil);
	NSParameterAssert(context != nil);

	NSFetchRequest *fetch = [[[NSFetchRequest alloc] init] autorelease];
	[fetch setEntity:[class entityInManagedObjectContext:context]];
	[fetch setPredicate:predicate];
	[fetch setSortDescriptors:sortDescriptors];
	NSArray *results = [context executeFetchRequest:fetch error:error];
	return results;
}

- (void)dispatchBlockWithMainContext:(void(^)(NSManagedObjectContext *context))block;
{
	NSParameterAssert(block != NULL);
	if (![NSThread isMainThread])
	{
		typeof(block) contextBlock = [block copy];
		dispatch_sync(dispatch_get_main_queue(), ^(void) {
			contextBlock(self.mainContext);
			[contextBlock release];
		});
	}
	else
	{
		block(self.mainContext);
	}
}

- (void)dispatchBlockWithManagedObjectContext:(void(^)(NSManagedObjectContext *context))block;
{
	NSParameterAssert(block != NULL);
	
	NSManagedObjectContext *context = [[NSManagedObjectContext alloc] init];
	id observation = [[[NSNotificationCenter defaultCenter] addObserverForName:NSManagedObjectContextDidSaveNotification object:context queue:nil usingBlock:^(NSNotification *note) {
		dispatch_async(dispatch_get_main_queue(), ^{
			[self.mainContext mergeChangesFromContextDidSaveNotification:note];
		});
	}] retain];
	[context setPersistentStoreCoordinator:self.storeCoordinator];
	block(context);
	[context release];
	[observation release];
}

- (void)dispatchAsyncBlockWithManagedObjectContext:(void(^)(NSManagedObjectContext *context))block;
{
	NSParameterAssert(block != NULL);
	
	typeof(block) contextBlock = [block copy];
	dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^(void) {
		NSManagedObjectContext *context = [[NSManagedObjectContext alloc] init];
		id observation = [[[NSNotificationCenter defaultCenter] addObserverForName:NSManagedObjectContextDidSaveNotification object:context queue:nil usingBlock:^(NSNotification *note) {
			dispatch_async(dispatch_get_main_queue(), ^{
				[self.mainContext mergeChangesFromContextDidSaveNotification:note];
			});
		}] retain];
		[context setPersistentStoreCoordinator:self.storeCoordinator];
		contextBlock(context);
		[contextBlock release];
		[context release];
		[observation release];
	});
}

@end
