//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1FileDetailsViewController.h"

#import "U1Logging.h"

#import "U1ByteSizeValueTransformer.h"
#import "U1FileNode.h"
#import "U1FilePreviewViewController.h"
#import "U1FilesClient.h"
#import "U1FuzzyDateTransformer.h"
#import "U1LocalFileInfo.h"
#import "U1Node.h"

enum U1FilePreviewActionSection {
	// TODO: these should have better names, but what...?
	U1FilePreviewActionSectionOne,
};

static const NSInteger U1FilePreviewActionSectionCount = 1;
static const NSInteger U1FilePreviewActionSectionOneFixedCount = 2;

enum U1FilePreviewActionSectionOneActions {
	U1FilePreviewActionView,
	U1FilePreviewActionPublish,
	U1FilePreviewActionPublishedStatus,
};


@interface U1FileDetailsViewController ()
@property (readwrite, retain) U1FileNode *node;
@property (retain) UIDocumentInteractionController *docController;
@property (copy) NSURL *contentURL;
@property (nonatomic, retain) U1LocalFileInfo *fileInfo;

NSDateFormatter *modificationDateFormatter(void);
U1ByteSizeValueTransformer *sizeTransformer(void);
U1FuzzyDateTransformer *dateTransformer(void);
- (void)reflowFileInfoView;
@end


@implementation U1FileDetailsViewController

@synthesize loadingContainerView, fileAncillaryDetailsView, iconImageView, nameLabel, loadingActivityView, sizeLabel, createdLabel, modifiedLabel, toolbar, publicURLLabel, renameTextField, sectionOneCells, sectionTwoCells, actionsTableView, publishSwitch, publishStatusLabel;
@synthesize node, docController, filesClient, contentURL, fileInfo;

NSDateFormatter *modificationDateFormatter(void)
{
	static NSDateFormatter *modificationDateFormatter = nil;
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		modificationDateFormatter = [[NSDateFormatter alloc] init];
		[modificationDateFormatter setDateStyle:NSDateFormatterMediumStyle];
		[modificationDateFormatter setTimeStyle:NSDateFormatterShortStyle];
	});
	return modificationDateFormatter;
}

U1ByteSizeValueTransformer *sizeTransformer(void)
{
	static U1ByteSizeValueTransformer *sizeTransformer = nil;
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		sizeTransformer = [[U1ByteSizeValueTransformer alloc] init];
	});
	return sizeTransformer;
}

U1FuzzyDateTransformer *dateTransformer(void)
{
	static U1FuzzyDateTransformer *dateTransformer = nil;
	static dispatch_once_t onceToken;
	dispatch_once(&onceToken, ^{
		dateTransformer = [[U1FuzzyDateTransformer alloc] init];
	});
	return dateTransformer;
}

- (id)initWithNode:(U1FileNode*)theNode;
{
	self = [super initWithNibName:@"U1FileDetailsViewController" bundle:nil];
	if (self == nil)
		return nil;
	filesClient = [[U1FilesClient sharedFilesClient] retain];
	self.node = theNode;
	self.title = [node.path lastPathComponent];
	self.hidesBottomBarWhenPushed = YES;
	self.navigationItem.rightBarButtonItem = [[[UIBarButtonItem alloc] initWithTitle:NSLocalizedString(@"Edit",@"")
																			   style:UIBarButtonItemStyleBordered
																			  target:self
																			  action:@selector(renameItem:)] autorelease];
	return self;
}

- (void)dealloc;
{
	[loadingContainerView release];
	[fileAncillaryDetailsView release];
	[iconImageView release];
	[nameLabel release];
	[loadingActivityView release];
	[sizeLabel release];
	[createdLabel release];
	[modifiedLabel release];
	[node release];
	[docController release];
	[contentURL release];
	[fileInfo release];
	[filesClient release];
	[toolbar release];
	[publicURLLabel release];
	[renameTextField release];
	[sectionOneCells release];
	[sectionTwoCells release];
	[actionsTableView release];
	[publishSwitch release];
	[publishStatusLabel release];
	[super dealloc];
}

- (void)setFileInfo:(U1LocalFileInfo *)value;
{
	if (value == fileInfo)
		return;
	
	[value retain];
	[fileInfo autorelease];
	fileInfo = value;
}

- (void)viewDidLoad;
{
	[super viewDidLoad];
	
	[self.nameLabel setText:self.title];
	NSString *sizeString = [sizeTransformer() transformedValue:[self.node size]];
	[self.sizeLabel setText:sizeString];
	NSString *dateString = [dateTransformer() transformedValue:self.node.modificationDate];
	[self.modifiedLabel setText:dateString];
	[self.publicURLLabel setText:[self.node publicURLString]];
	[self.publishSwitch setOn:[self.node isPublicValue]];
	[self.publishStatusLabel setText:[self.node publicURLString]];
	self.fileInfo = [self.filesClient localInfoForNode:self.node];
    self.docController = [UIDocumentInteractionController interactionControllerWithURL:fileInfo.fileURL];
	[self.iconImageView setImage:[self.docController.icons lastObject]];
	// Set the fileAncillaryDetailsView's frame's origin.Y to be self.nameLabel's frame's height + 1
	[self reflowFileInfoView];
}

- (void)reflowFileInfoView;
{
    CGSize expectedLabelSize = [self.title sizeWithFont:self.nameLabel.font 
									  constrainedToSize:self.nameLabel.frame.size
										  lineBreakMode:UILineBreakModeWordWrap];
	
	[self.fileAncillaryDetailsView setFrame:CGRectMake(self.fileAncillaryDetailsView.frame.origin.x,
													   (self.nameLabel.frame.origin.y + expectedLabelSize.height + 1), 
													   self.fileAncillaryDetailsView.frame.size.width,
													   self.fileAncillaryDetailsView.frame.size.height)];	
}

- (void)viewDidUnload;
{
	[loadingContainerView release], loadingContainerView = nil;
	[iconImageView release], iconImageView = nil;
	[nameLabel release], nameLabel = nil;
	[loadingActivityView release], loadingActivityView = nil;
	[sizeLabel release], sizeLabel = nil;
	[createdLabel release], createdLabel = nil;
	[modifiedLabel release], modifiedLabel = nil;
	[toolbar release], toolbar = nil;
	[publicURLLabel release], publicURLLabel = nil;
	[renameTextField release], renameTextField = nil;
	[sectionOneCells release], sectionOneCells = nil;
	[sectionTwoCells release], sectionTwoCells = nil;
	[actionsTableView release], actionsTableView = nil;
	[publishSwitch release], publishSwitch = nil;
	[publishStatusLabel release], publishStatusLabel = nil;
	[super viewDidUnload];
}

- (void)viewDidAppear:(BOOL)animated;
{
	NSIndexPath *selectedIndexPath = [self.actionsTableView indexPathForSelectedRow];
	[self.actionsTableView deselectRowAtIndexPath:selectedIndexPath animated:YES];
}

- (void)setEditing:(BOOL)editing animated:(BOOL)animated;
{
	[super setEditing:editing animated:animated];
	if (editing)
	{
		[self.renameTextField setText:self.title];
		[self.renameTextField setHidden:NO];
		[self.nameLabel setHidden:YES];
		[self.renameTextField becomeFirstResponder];
		
	}
	else
	{
		[self.renameTextField setHidden:YES];
		[self.nameLabel setHidden:NO];
	}
}

- (BOOL)canBecomeFirstResponder;
{
	return YES;
}

- (IBAction)publishItem:(UISwitch*)sender;
{
	[sender setEnabled:NO];
	
	NSArray *publishStatusIndexPath = [NSArray arrayWithObject:[NSIndexPath indexPathForRow:U1FilePreviewActionPublishedStatus inSection:U1FilePreviewActionSectionOne]];
	
	if (![sender isOn] && [self.node isPublicValue])
	{
		[self.publishStatusLabel setText:NSLocalizedString(@"Unpublishing file...", nil)];
		[self.filesClient unpublishNode:self.node completionBlockOnMainQueue:^(U1FileNode *node, NSError *error) {
			[sender setEnabled:YES];
			[self.publishStatusLabel setText:nil];
			[self.actionsTableView deleteRowsAtIndexPaths:publishStatusIndexPath withRowAnimation:UITableViewRowAnimationFade];
		}];
	}
	else if ([sender isOn] && ![self.node isPublicValue])
	{
		[self.actionsTableView beginUpdates];
		[self.publishStatusLabel setText:NSLocalizedString(@"Publishing file...", nil)];
		[self.actionsTableView insertRowsAtIndexPaths:publishStatusIndexPath withRowAnimation:UITableViewRowAnimationNone];
		[self.actionsTableView endUpdates];
		[self.filesClient publishNode:self.node completionBlockOnMainQueue:^(U1FileNode *node, NSError *error) {
			[sender setEnabled:YES];
			[self.actionsTableView beginUpdates];
			[self.publishStatusLabel setText:[self.node publicURLString]];
			[self.actionsTableView reloadRowsAtIndexPaths:publishStatusIndexPath withRowAnimation:UITableViewRowAnimationNone];
			[self.actionsTableView endUpdates];
		}];
	}
}

- (IBAction)renameItem:(UIBarButtonItem*)sender;
{
	[self setEditing:YES animated:YES];
}

- (IBAction)deleteItem:(UIBarButtonItem*)sender;
{
	UIActionSheet *actionSheet = [[UIActionSheet alloc] initWithTitle:NSLocalizedString(@"Are you sure you want to delete this file?\nIt will be deleted everywhere in the Ubuntu One service.", @"")
															 delegate:self
													cancelButtonTitle:NSLocalizedString(@"Cancel", @"")
											   destructiveButtonTitle:NSLocalizedString(@"Delete", @"")
													otherButtonTitles:nil];
	[actionSheet showFromToolbar:self.toolbar];
	[actionSheet release];
}

- (void)actionSheet:(UIActionSheet *)actionSheet clickedButtonAtIndex:(NSInteger)buttonIndex;
{
	if (buttonIndex == 0)
	{
		[self.filesClient deleteNode:self.node completionBlock:^(NSError *error) {
			if (error) {
				CLS_LOG(@"error in deleting a node: %@", error);
			}
		}];
		[self.navigationController popViewControllerAnimated:YES];
	}
}

//- (IBAction)downloadItem:(UIBarButtonItem*)sender;
//{
//	self.fileInfo = [self.filesClient downloadFileForNode:self.node completionBlock:^(NSError *error) {
//		[self.fileInfo removeObserver:self forKeyPath:@"downloadPercentage"];
//		[self.fileInfo removeObserver:self forKeyPath:@"downloading"];
//		[self.fileInfo removeObserver:self forKeyPath:@"downloaded"];
//	}];
//	[self.fileInfo addObserver:self forKeyPath:@"downloadPercentage" options:0 context:self.fileInfo];
//	[self.fileInfo addObserver:self forKeyPath:@"downloading" options:NSKeyValueObservingOptionInitial context:self.fileInfo];
//	[self.fileInfo addObserver:self forKeyPath:@"downloaded" options:0 context:self.fileInfo];
//}
//
//- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context;
//{
//	if (context == self.fileInfo && [keyPath isEqualToString:@"downloadPercentage"])
//	{
//	}
//	else if (context == self.fileInfo && [keyPath isEqualToString:@"downloading"])
//	{
//		if ([self.fileInfo isDownloading])
//		{
//		}
//	}
//	else if (context == self.fileInfo && [keyPath isEqualToString:@"downloaded"])
//	{
//		if ([self.fileInfo isDownloaded])
//		{
//		}
//	}
//	else
//	{
//		[super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
//	}
//}

- (IBAction)showPreview:(UIBarButtonItem*)sender;
{
	
	U1FilePreviewViewController *previewController = [[U1FilePreviewViewController alloc] initWithNode:self.node];
	[self.navigationController pushViewController:previewController animated:YES];
	[previewController release];
	
//	QLPreviewController *previewController = [[QLPreviewController alloc] init];
//	previewController.dataSource = self;
//	previewController.delegate = self;
//
//	UILabel *titleLabel = [[UILabel alloc] initWithFrame:CGRectMake(0, 0, 100, 20)];
//	// TODO: Ubuntu font
//	[titleLabel setFont:[UIFont boldSystemFontOfSize:18.f]];
//	[titleLabel setTextColor:[UIColor blackColor]];
//	[titleLabel setShadowColor:[UIColor whiteColor]];
//	[titleLabel setShadowOffset:CGSizeMake(-1, -1)];
//	[titleLabel setText:self.title];
//	[titleLabel sizeToFit];
//	[titleLabel setOpaque:NO];
//	[titleLabel setBackgroundColor:[UIColor clearColor]];
//	[previewController.navigationItem setTitleView:titleLabel];
//	[titleLabel release];
//
////	[self.navigationController presentModalViewController:previewController animated:YES];
//	[self.navigationController pushViewController:previewController animated:YES];
//	[previewController release];
//
//	// TODO: manage preview view vs. modal presentation
////	UIView *previewView = previewController.view;
////	previewView.frame = [self.view bounds];
////	[previewController viewWillAppear:NO];
////	[self.view addSubview:previewView];
////	[previewController viewDidAppear:NO];
}

- (IBAction)openIn:(UIBarButtonItem*)sender;
{
	BOOL success = [self.docController presentOpenInMenuFromBarButtonItem:sender animated:YES];
	if(!success){
		UIAlertView *alertView = [[UIAlertView alloc] initWithTitle:@"Unable to Open File"
															message:@"No apps that can open this file type were found."
														   delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil];
		[alertView show];
	}
}

- (IBAction)copyPublicURL:(id)sender;
{
	[[UIPasteboard generalPasteboard] setURL:[NSURL URLWithString:[self.node publicURLString]]];
}

- (IBAction)openPublicURL:(id)sender;
{
	NSURL *publicURL = [NSURL URLWithString:[self.node publicURLString]];
	[[UIApplication sharedApplication] openURL:publicURL];
}


#pragma mark UITextFieldDelegate Methods

- (BOOL)textFieldShouldReturn:(UITextField *)textField;
{
	[textField endEditing:NO];
//	[textField resignFirstResponder];
	return YES;
}

- (void)textFieldDidEndEditing:(UITextField *)textField;
{
	NSString *newName = [self.renameTextField text];
	
	[self setEditing:NO animated:YES];
	
	if (!newName)
		return;

	[[U1FilesClient sharedFilesClient] renameNode:self.node toName:newName completionBlockOnMainQueue:^(U1Node *newNode, NSError *error) {
		
		if (!error)
		{
			self.title = [newNode.path lastPathComponent];
			UILabel *titleLabel = (UILabel*)[self.navigationItem titleView];
			[titleLabel setText:self.title];
			[titleLabel sizeToFit];
			[self.nameLabel setText:self.title];
			[self reflowFileInfoView];
		}
	}];
}


#pragma mark UITableViewDataSource Methods

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView;
{
	return U1FilePreviewActionSectionCount;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section;
{
	switch (section)
	{
		case U1FilePreviewActionSectionOne:
			return ([self.node isPublicValue] || [self.publishSwitch isOn]) ? U1FilePreviewActionSectionOneFixedCount + 1 : U1FilePreviewActionSectionOneFixedCount;
			break;
	}
	return 0;
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath;
{
	switch (indexPath.section)
	{
		case U1FilePreviewActionSectionOne:
			return [self.sectionOneCells objectAtIndex:indexPath.row];
			break;
		}
	return nil;
}


#pragma mark UITableViewDelegate Methods

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath;
{
	switch (indexPath.section)
	{
		case U1FilePreviewActionSectionOne:
		{
			switch (indexPath.row)
			{
				case U1FilePreviewActionView:
					[self showPreview:nil];
					break;
					
				case U1FilePreviewActionPublishedStatus:
				{
					UIMenuController *menuController = [UIMenuController sharedMenuController];
					menuController.menuItems = [NSArray arrayWithObjects:
												[[[UIMenuItem alloc] initWithTitle:NSLocalizedString(@"Copy", nil) action:@selector(copyPublicURL:)] autorelease],
												[[[UIMenuItem alloc] initWithTitle:NSLocalizedString(@"Open", nil) action:@selector(openPublicURL:)] autorelease],
												nil];
					
					[menuController setTargetRect:[self.actionsTableView rectForRowAtIndexPath:indexPath] inView:self.actionsTableView];
					[menuController setMenuVisible:YES animated:YES];
					
					__block id observer = nil;
					observer = [[[NSNotificationCenter defaultCenter] addObserverForName:UIMenuControllerDidHideMenuNotification object:menuController queue:nil usingBlock:^(NSNotification *note) {
						[self.actionsTableView deselectRowAtIndexPath:indexPath animated:YES];
						[observer release], observer = nil;
					}] retain];
					break;
				}
			
			}
			break;
		}
			

	}
	[self.actionsTableView deselectRowAtIndexPath:indexPath animated:YES];
}

@end
