//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1FilesClient.h"

#import <CoreData/CoreData.h>

#import "U1Logging.h"

#import "Reachability.h"

#import "U1AccountManager.h"
#import "U1Asset.h"
#import "U1DataRepository.h"
#import "U1FilesService.h"
#import "U1FileNode.h"
#import "U1FolderNode.h"
#import "U1AutoUploadsManager.h"
#import "U1LocalFileInfo.h"
#import "U1Node.h"
#import "U1UploadOperation.h"
#import "U1Volume.h"


NSString * const U1FilesClientNodeStatusChangedNotification = @"U1FilesClientNodeStatusChangedNotification";


@interface U1FilesClient ()
@property (retain) NSOperationQueue *uploadsQueue;
- (NSURL*)fileURLForNode:(U1FileNode*)node;
@end

static U1FilesClient *sharedFilesClient = nil;

@implementation U1FilesClient

@synthesize dataRepository, filesService;
@synthesize uploadsQueue;

+ (U1FilesClient*)sharedFilesClient;
{
	if (sharedFilesClient == nil)
	{
		sharedFilesClient = [[self alloc] init];
	}
	return sharedFilesClient;
}

- (id)init;
{
	self = [super init];
	if (self == nil)
		return nil;
	dataRepository = [U1DataRepository sharedDataRepository];
	filesService = [U1FilesService sharedFilesService];
	uploadsQueue = [[NSOperationQueue alloc] init];
	[self.uploadsQueue setMaxConcurrentOperationCount:1];
	NSNotificationCenter *defaultCenter = [NSNotificationCenter defaultCenter];
	[defaultCenter addObserver:self selector:@selector(reachabilityChanged:) name:kReachabilityChangedNotification object:nil];
	[defaultCenter addObserver:self selector:@selector(didSignOut:) name:U1AccountManagerDidSignOut object:nil];
	return self;
}

- (void)dealloc;
{
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	[uploadsQueue cancelAllOperations];
	[uploadsQueue release];
	[super dealloc];
}

- (void)reachabilityChanged:(NSNotification*)notification;
{
	Reachability *reachability = [notification object];
	NetworkStatus status = [reachability currentReachabilityStatus];
	switch (status)
	{
		case ReachableViaWiFi:
			[self.uploadsQueue setSuspended:NO];
			break;
		case ReachableViaWWAN:
		{
			[self.uploadsQueue setSuspended:NO];
			break;
		}
		case NotReachable:
			[self.uploadsQueue setSuspended:YES];
			break;
	}
}

- (void)didSignOut:(NSNotification*)notification;
{
	[self.uploadsQueue cancelAllOperations];
}

- (id)accountInfoWithCompletionBlock:(void(^)(NSDictionary *accountInfo, NSError *error))completionBlock;
{
	return [self.filesService accountInfoWithCompletionBlockOnMainQueue:completionBlock];
}

- (id)volumesWithResultsBlock:(void(^)(U1Volume *rootVolume, NSFetchedResultsController *cloudFoldersResultsController, U1DataSourceType resultsDataSourceType))resultsBlock errorBlock:(void(^)(NSError *error))errorBlock;
{
	NSParameterAssert(resultsBlock != NULL);
	NSParameterAssert(errorBlock != NULL);
	
	__block NSError *error = nil;
	__block NSFetchedResultsController *cloudFoldersResultsController = nil;
	__block U1Volume *rootVolume = nil; // vestigial
	
	[self.dataRepository dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
		
		NSFetchRequest *request = [[NSFetchRequest alloc] init];
		[request setEntity:[U1Volume entityInManagedObjectContext:context]];
		NSArray *sortDescriptors = [NSArray arrayWithObjects:
									[NSSortDescriptor sortDescriptorWithKey:@"kind" ascending:YES],
									[NSSortDescriptor sortDescriptorWithKey:@"resourcePath" ascending:YES selector:@selector(localizedStandardCompare:)],
									nil];
		[request setSortDescriptors:sortDescriptors];
		cloudFoldersResultsController = [[NSFetchedResultsController alloc] initWithFetchRequest:request managedObjectContext:context sectionNameKeyPath:nil cacheName:nil];
		[request release];
	}];
	
	BOOL success = [cloudFoldersResultsController performFetch:&error];
	if (!success)
	{
        
		[cloudFoldersResultsController release];
		errorBlock(error);
		return nil;
	}
	
	resultsBlock(rootVolume, cloudFoldersResultsController, U1DataSourceLocal);
	
	[self.filesService volumeInfoWithCompletionBlockOnMainQueue:^(NSArray *volumeInfos, NSError *error) {
		
		[self.dataRepository dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
			
			NSMutableArray *fetchedVolumes = [NSMutableArray array];
			for (NSDictionary *volumeInfo in volumeInfos)
			{
				NSString *path = [volumeInfo objectForKey:@"path"];
				NSPredicate *p = [NSPredicate predicateWithFormat:@"path = %@", path];
				U1Volume *volume = [[self.dataRepository resultsForEntityClass:[U1Volume class] matchingPredicate:p withSortDescriptors:nil error:NULL] lastObject];
				if (!volume)
				{
					volume = [U1Volume insertInManagedObjectContext:context];
				}
				[volume updatePropertiesFromJSONDictionary:volumeInfo];
				if ([volume.kind isEqualToString:@"root"])
				{
					rootVolume = [volume retain];
				}
				[fetchedVolumes addObject:volume];
			}
			
			NSPredicate *p = [NSPredicate predicateWithFormat:@"not (self in %@)", fetchedVolumes];
			NSArray *deadVolumes = [self.dataRepository resultsForEntityClass:[U1Volume class] matchingPredicate:p withSortDescriptors:nil error:NULL];
			for (U1Volume *deadVolume in deadVolumes)
			{
				[context deleteObject:deadVolume];
			}

			[context save:NULL];
		}];
		
		dispatch_async(dispatch_get_main_queue(), ^(void) {
			if (error)
				errorBlock(error);
			else
			{
				resultsBlock(rootVolume, cloudFoldersResultsController, U1DataSourceRemote);
				[[NSNotificationCenter defaultCenter] postNotificationName:@"volumesFetched" object:[cloudFoldersResultsController fetchedObjects]];
			}
			[rootVolume release];
			[cloudFoldersResultsController release];
		});
	}];
	return nil;
}

- (id)fetchFolderWithResourcePath:(NSString*)nodePath resultsBlock:(void(^)(U1Node *node, NSFetchedResultsController *childrenResultsController, U1DataSourceType resultsDataSourceType))resultsBlock errorBlock:(void(^)(NSError *error))errorBlock;
{

	__block U1FolderNode *node = [(id)[self.dataRepository nodeWithResourcePath:nodePath] retain];
	__block NSFetchedResultsController *resultsController = nil;

	[self.dataRepository dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {

		NSError *err = nil;
        
		if (!node)
		{
			CLS_LOG(@"no node for path %@, creating one.", nodePath);
			node = [[U1FolderNode insertInManagedObjectContext:context] retain];
			node.resourcePath = nodePath;
			BOOL success = [context save:&err];
			if(!success){
				CLS_LOG(@"error saving new U1FolderNode for path: %@ err: %@", nodePath, err);
			}
		}
		
		NSFetchRequest *request = [[NSFetchRequest alloc] init];
		[request setEntity:[U1Node entityInManagedObjectContext:context]];
		NSPredicate *p = [NSPredicate predicateWithFormat:@"parent = %@", node];
		[request setPredicate:p];
		NSArray *sortDescriptors = [NSArray arrayWithObjects:
									[NSSortDescriptor sortDescriptorWithKey:@"kind" ascending:YES],
									[NSSortDescriptor sortDescriptorWithKey:@"resourcePath" ascending:YES selector:@selector(localizedStandardCompare:)],
									nil];
		[request setSortDescriptors:sortDescriptors];
		resultsController = [[NSFetchedResultsController alloc] initWithFetchRequest:request managedObjectContext:context sectionNameKeyPath:nil cacheName:nil];
		[request release];
	}];

	NSError *error = nil;
	BOOL success = [resultsController performFetch:&error];
	if (!success)
	{
		CLS_LOG(@"Error in local fetch for children of node");
		errorBlock(error);
		[resultsController release];
		[node release];
		return nil;
	}
	
	resultsBlock(node, resultsController, U1DataSourceLocal);
	
	[self.filesService infoForNodeAtResourcePath:node.resourcePath includeChildren:YES completionBlockOnMainQueue:^(NSDictionary *nodeInfo, NSError *error) {

		if (error)
		{
			CLS_LOG(@"error from call to filesService infoForNodeAtResourcePath:. %@", error);
			errorBlock(error);
			return;
		}
		
		[node updatePropertiesFromJSONDictionary:nodeInfo];
		
		NSManagedObjectID *nodeID = [node objectID];
		
		NSArray *childInfos = [nodeInfo objectForKey:@"children"];
		[self.dataRepository dispatchAsyncBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {

			NSError *error = nil;

			U1FolderNode *folderNode = (id)[context objectWithID:nodeID];
			NSMutableSet *orphans = [NSMutableSet set];
			[orphans unionSet:folderNode.children];
			
			NSUInteger batchCounter = 0;
			for (NSDictionary *childInfo in childInfos)
			{
				// Ignore nodes with no contentblob:
				if ([childInfo objectForKey:@"size"] == [NSNull null]) {
					continue;
				}
			
				NSString *childResourcePath = [childInfo objectForKey:@"resource_path"];
				NSString *kind = [childInfo objectForKey:@"kind"];
				NSPredicate *p = [NSPredicate predicateWithFormat:@"resourcePath = %@", childResourcePath];
				NSArray *results = [self.dataRepository resultsForEntityClass:[U1Node class] inContext:context withPredicate:p sortDescriptors:nil error:&error];
				if(results == nil) {
					CLS_LOG(@"error in getting nodes for path '%@' err=%@", childResourcePath, error);
				}
				U1Node *childNode = [results lastObject];
				if (!childNode)
				{
					if ([@"file" isEqualToString:kind])
						childNode = [U1FileNode insertInManagedObjectContext:context];
					else
						childNode = [U1FolderNode insertInManagedObjectContext:context];
				}
                
				[childNode updatePropertiesFromJSONDictionary:childInfo];
				childNode.parent = folderNode;
				[orphans removeObject:childNode];
				
				if (++batchCounter > 100)
				{
					if(![context save:&error])
					{
						CLS_LOG(@"error saving child node batch: %@", error);
					}
					batchCounter = 0;
				}
			}

			// save last batch of new nodes before attempting deletes:
			if(![context save:&error]){
				CLS_LOG(@"error saving context: %@", error);
			}

			for (U1Node *orphan in orphans)
			{
				if (orphan.generation != nil) // Only delete objects that existed at one time.
				{
					[context deleteObject:orphan];
				}
			}
			
			if(![context save:&error]){
				CLS_LOG(@"error saving context: %@", error);
			}

			dispatch_async(dispatch_get_main_queue(), ^(void) {
				resultsBlock(node, resultsController, U1DataSourceRemote);
				[resultsController release];
				[node release];
			});
		
		}];

		
	}];
	return nil;
}

- (id)fetchFileWithResourcePath:(NSString *)nodePath completionBlock:(void (^)(U1Node *, NSError *))completionBlock;
{
	__block U1FileNode *node = [(id)[self.dataRepository nodeWithResourcePath:nodePath] retain];
	
	if (!node)
	{
		[self.dataRepository dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
			NSError *error = nil;
            
			node = [[U1FileNode insertInManagedObjectContext:context] retain];
			node.resourcePath = nodePath;
			if(![context save:&error]){
				CLS_LOG(@"error saving new file node at path '%@'", nodePath);
			}
		}];
	}
	
	
	[self.filesService infoForNodeAtResourcePath:node.resourcePath includeChildren:NO completionBlockOnMainQueue:^(NSDictionary *nodeInfo, NSError *error) {
		if (error == nil)
		{
			[node updatePropertiesFromJSONDictionary:nodeInfo];
		}
		dispatch_async(dispatch_get_main_queue(), ^(void) {
			NSError *error = nil;
			completionBlock(node, error);
			[node release];
		});
	}];


	return nil;
}

- (id)renameNode:(U1Node*)node toName:(NSString*)newName completionBlockOnMainQueue:(void(^)(U1Node *node, NSError *error))completionBlock;
{
	NSParameterAssert(node != nil);
	NSParameterAssert(newName != nil);
	NSParameterAssert(completionBlock != NULL);
	
	NSString *newPath = [[node.path stringByDeletingLastPathComponent] stringByAppendingPathComponent:newName];
	return [self.filesService moveNodeAtResourcePath:node.resourcePath toPath:newPath withCompletionBlockOnMainQueue:^(NSDictionary *nodeInfo, NSError *error) {
		if (!error)
		{
			[node updatePropertiesFromJSONDictionary:nodeInfo];
		}
		completionBlock(node, error);
	}];
}

- (id)deleteNode:(U1Node*)node completionBlock:(void(^)(NSError *error))completionBlock;
{
	return [self.filesService deleteNodeAtResourcePath:node.resourcePath completionBlockOnMainQueue:^(NSError *error) {
		if (!error)
		{
			[self.dataRepository dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
				[context deleteObject:node];
				[context save:NULL];
			}];
		}
		completionBlock(error);
	}];
}

- (id)uploadAssetWithID:(NSManagedObjectID*)assetID withPriority:(NSOperationQueuePriority)priority progressBlock:(void (^)(long long, long long))progressBlock completionBlockOnMainQueue:(void (^)(U1FileNode *, NSError *))completionBlock;
{

	
	U1UploadOperation *operation = [[U1UploadOperation alloc] init];
	__block typeof(operation) operationRef = [operation retain];

	operation.assetID = assetID;
	operation.progressBlock = ^(long long sent, long long total) {
		
		dispatch_async(dispatch_get_main_queue(), ^{

			U1Asset *asset = (U1Asset *)[[self.dataRepository mainContext] objectWithID:assetID];

			NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
									  asset.fileNode, @"node",
									  operationRef.fileInfo, @"fileInfo",
									  nil];
			[[NSNotificationCenter defaultCenter] postNotificationName:U1FilesClientNodeStatusChangedNotification object:self userInfo:userInfo];
			
		});

		if (progressBlock)
		{
			progressBlock(sent, total);
		}
	};
	[operation setQueuePriority:priority];
	[operation setCompletionBlock:^{
		dispatch_async(dispatch_get_main_queue(), ^(void) {
			
			if (operationRef.error == nil) // or the error is somehow unrecoverable (e.g., over quota)
			{
				U1Asset *asset = (U1Asset *)[[self.dataRepository mainContext] objectWithID:assetID];

				[[NSNotificationCenter defaultCenter] postNotificationName:@"imageUploaded" object:nil];
				completionBlock(asset.fileNode, nil);
			}
			else
			{
				NSLog(@"Upload operation failed both magic and regular uploads: %@", operationRef.error);
				completionBlock(nil, operationRef.error);
			}
            [operationRef release];
		});
	}];
	
	[self.uploadsQueue addOperation:operation];
	[[NSNotificationQueue defaultQueue] enqueueNotification:[NSNotification notificationWithName:@"uploadQueued"
																						  object:nil]
											   postingStyle:NSPostWhenIdle
											   coalesceMask:NSNotificationCoalescingOnName
												   forModes:nil];
	[operation release];
	return operation;
}

- (id)publishNode:(U1FileNode*)node completionBlockOnMainQueue:(void(^)(U1FileNode *node, NSError *error))completionBlock;
{
	return [self.filesService publishNode:node completionBlockOnMainQueue:completionBlock];
}

- (id)unpublishNode:(U1FileNode*)node completionBlockOnMainQueue:(void(^)(U1FileNode *node, NSError *error))completionBlock;
{
	return [self.filesService unpublishNode:node completionBlockOnMainQueue:completionBlock];
}

- (id)createFolderNamed:(NSString*)name inFolder:(U1FolderNode*)folderNode completionBlockOnMainQueue:(void(^)(U1FolderNode *newFolderNode, NSError *error))completionBlock;
{
	NSString *resourcePath = folderNode.resourcePath;
	resourcePath = [resourcePath stringByAppendingPathComponent:name];

	U1FolderNode *node = (id)[self.dataRepository nodeWithResourcePath:resourcePath];
	if (!node)
	{
		[self.dataRepository dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
			U1FolderNode *newNode = [U1FolderNode insertInManagedObjectContext:context];
			newNode.resourcePath = resourcePath;
			newNode.kind = @"directory";
			newNode.contentPath = [folderNode.contentPath stringByAppendingPathComponent:name];
			newNode.parent = folderNode;
			[context save:NULL];
		}];
		node = (id)[self.dataRepository nodeWithResourcePath:resourcePath];
	}
	
	return [self.filesService createFolderAtPath:resourcePath completionBlockOnMainQueue:^(NSDictionary *nodeInfo, NSError *error) {
		if (!error)
		{
			[node updatePropertiesFromJSONDictionary:nodeInfo];
		}
		completionBlock(node, error);
	}];
}

- (id)createVolumeAtPath:(NSString*)folderPath completionBlockOnMainQueue:(void(^)(U1Volume *volume, NSError *error))completionBlock;
{
	return [self.filesService createVolumeAtPath:folderPath completionBlockOnMainQueue:^(NSDictionary *volumeInfo, NSError *error) {
		[self.dataRepository dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
			U1Volume *volume = [U1Volume insertInManagedObjectContext:context];
			if (!error)
			{
				[volume updatePropertiesFromJSONDictionary:volumeInfo];
			}
			completionBlock(volume, error);
		}];
	}];
}

- (BOOL)isNodeSavedLocally:(U1FileNode*)node;
{
	NSString *contentPath = node.contentPath;
	
	NSArray *paths = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
	NSString *documentsDirectory = [paths objectAtIndex:0];
	
	NSString *filePath = [documentsDirectory stringByAppendingPathComponent:contentPath];
	NSFileManager *fileManager = [[[NSFileManager alloc] init] autorelease];
	return [fileManager fileExistsAtPath:filePath];
}

- (U1LocalFileInfo*)localInfoForNode:(U1FileNode*)node;
{
	// TODO: cache info
	U1LocalFileInfo *info = [[U1LocalFileInfo new] autorelease];
	info.fileURL = [self fileURLForNode:node];
	NSFileManager *fileManager = [[[NSFileManager alloc] init] autorelease];
	info.downloaded = [fileManager fileExistsAtPath:[info.fileURL path]];
	info.downloading = NO;
	return info;
}

- (NSURL*)fileURLForNode:(U1FileNode*)node;
{
	NSString *contentPath = node.contentPath;
	NSArray *paths = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
	NSString *documentsDirectory = [paths objectAtIndex:0];
	NSString *filePath = [documentsDirectory stringByAppendingPathComponent:contentPath];
	NSURL *fileURL = [NSURL fileURLWithPath:filePath];
	return fileURL;
}

- (U1LocalFileInfo*)downloadFileForNode:(U1FileNode*)node completionBlock:(void(^)(NSError *error))completionBlock;
{
	U1LocalFileInfo *info = [[U1LocalFileInfo new] autorelease];
	info.fileURL = [self fileURLForNode:node];
	info.downloaded = NO;
	info.downloading = YES;
	info.downloadPercentage = 0.f;

	CGFloat expected = [[node size] longLongValue];
	
	[self.filesService contentForNode:node progressBlock:^(long long bytesReceived, long long bytesExpected) {
		
		dispatch_async(dispatch_get_main_queue(), ^(void) {

			CGFloat received = (CGFloat)bytesReceived;
			info.downloadPercentage = (received / expected);
			
		});

	} completionBlock:^(NSURL *contentURL) {
		dispatch_async(dispatch_get_main_queue(), ^(void) {
			
			info.downloadPercentage = 100.f;
			info.downloading = NO;
			info.downloaded = YES;
			
		});
	}];

	return info;
}

- (BOOL)isOperating;
{
	return ([self queueCount] > 0 && ![self.uploadsQueue isSuspended]);
}

- (BOOL)isPending:(U1FileNode *)fileNode;
{
	__block BOOL isPending = NO;
	[self.dataRepository dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {
		
		[[self.uploadsQueue operations] enumerateObjectsUsingBlock:^(U1UploadOperation *operation, NSUInteger idx, BOOL *stop) {
			U1Asset *asset = (U1Asset *)[context objectWithID:operation.assetID];
			if (asset.fileNode == fileNode)
			{
				isPending = YES;
				*stop = YES;
			}
		}];
	}];
	
	return isPending;
}

- (BOOL)isUploading:(U1FileNode *)fileNode;
{
	__block BOOL isUploading = NO;
	[self.dataRepository dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {
		
		[[self.uploadsQueue operations] enumerateObjectsUsingBlock:^(U1UploadOperation *operation, NSUInteger idx, BOOL *stop) {
			U1Asset *asset = (U1Asset *)[context objectWithID:operation.assetID];
			
			if (asset.fileNode == fileNode && [operation isExecuting])
			{
				isUploading = YES;
				*stop = YES;
			}
		}];
	}];
		
	return isUploading;
}

- (U1LocalFileInfo*)fileInfoForNode:(U1FileNode*)fileNode;
{
	__block U1UploadOperation *uploadOperation = nil;
	[self.dataRepository dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {
		
		[[self.uploadsQueue operations] enumerateObjectsUsingBlock:^(U1UploadOperation *operation, NSUInteger idx, BOOL *stop) {
			U1Asset *asset = (U1Asset *)[context objectWithID:operation.assetID];

			if (asset.fileNode == fileNode)
			{
				uploadOperation = operation;
				*stop = YES;
			}
		}];
	}];
	return uploadOperation.fileInfo;
}

- (int)queueCount;
{
	U1AutoUploadsManager *sharedAutoUploadsManager = [U1AutoUploadsManager sharedAutoUploadsManager];
	return [sharedAutoUploadsManager pendingCount] + [self.uploadsQueue operationCount];
}
@end
