//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1FilesService.h"

#import "OAuthConsumer.h"
#import "Reachability.h"

#import "U1Logging.h"

#import "U1AccountManager.h"
#import "U1Asset.h"
#import "U1DataRepository.h"
#import "U1FileNode.h"
#import "U1HTTPRequestOperation.h"
#import "U1Node.h"
#import "U1ReportingInputStream.h"

static U1FilesService *sharedFilesService = nil;


@interface U1FilesService ()

@property (retain) NSOperationQueue *networkQueue;
@property (retain) Reachability *reachability;

- (NSURL*)URLForVolumeWithPath:(NSString*)volumePath;
- (NSURL*)URLForNode:(NSString*)nodePath includeChildren:(BOOL)includeChildren;
- (NSURL*)contentURLForNode:(NSString*)nodePath;
- (id)execute:(NSString*)method toURL:(NSURL*)url requestBody:(NSString*)body parseResponseBody:(BOOL)parseResponseBody completionBlockOnMainQueue:(void(^)(id results, NSError *error))completionBlock;
- (id)execute:(NSString*)method toURL:(NSURL*)url requestBody:(id)bodyData contentType:(NSString*)contentType contentLength:(NSUInteger)contentLength parseResponseBody:(BOOL)parseResponseBody completionBlockOnMainQueue:(void(^)(id results, NSError *error))completionBlock;

@end


@implementation U1FilesService

static NSString *const U1FilesServiceAPIBase = @"https://one.ubuntu.com";
static NSString *const U1FilesServiceAPIPath = @"/api/file_storage";
static NSString *const U1FilesServiceAPIVersion1 = @"/v1";
static NSString *const U1FilesServiceAPIContentRoot = @"https://files.one.ubuntu.com/content";
static NSString *const U1ReachabilityHost = @"files.one.ubuntu.com";

@synthesize networkQueue, reachability;

+ (U1FilesService*)sharedFilesService;
{
	if (sharedFilesService == nil)
	{
		sharedFilesService = [[self alloc] init];
	}
	return sharedFilesService;
}

- (id)init;
{
	self = [super init];
	if (self == nil)
		return nil;
	networkQueue = [[NSOperationQueue alloc] init];
	reachability = [[Reachability reachabilityWithHostName:U1ReachabilityHost] retain];
	[reachability startNotifer];
	return self;
}

- (void)dealloc;
{
	[networkQueue release];
	[reachability stopNotifer];
	[reachability release];
	[super dealloc];
}

- (id)accountInfoWithCompletionBlockOnMainQueue:(void(^)(NSDictionary *accountInfo, NSError *error))completionBlock;
{
	NSMutableString *urlString = [NSMutableString stringWithString:U1FilesServiceAPIBase];
	[urlString appendString:U1FilesServiceAPIPath];
	[urlString appendString:U1FilesServiceAPIVersion1];
	NSURL *rootURL = [NSURL URLWithString:urlString];
	return [self execute:@"GET" toURL:rootURL requestBody:nil parseResponseBody:YES completionBlockOnMainQueue:^(id results, NSError *error) {
		completionBlock(results, error);
	}];
}

- (id)volumeInfoWithCompletionBlockOnMainQueue:(void(^)(NSArray *volumeInfos, NSError *error))completionBlock;
{
	NSURL *volumesURL = [self URLForNode:@"/volumes" includeChildren:NO];
	return [self execute:@"GET" toURL:volumesURL requestBody:nil parseResponseBody:YES completionBlockOnMainQueue:^(id results, NSError *error) {
		completionBlock(results, error);
	}];
}

- (id)infoForNodeAtResourcePath:(NSString*)resourcePath includeChildren:(BOOL)includeChildren completionBlockOnMainQueue:(void(^)(NSDictionary *node, NSError *error))completionBlock;
{
	NSURL *nodeURL = [self URLForNode:resourcePath includeChildren:includeChildren];

	return [self execute:@"GET" toURL:nodeURL requestBody:nil parseResponseBody:YES completionBlockOnMainQueue:^(id result, NSError *error) {
		
		completionBlock(result, error);
	}];
}

- (id)infoForNode:(U1Node*)node includeChildren:(BOOL)includeChildren completionBlockOnMainQueue:(void(^)(NSDictionary *node, NSError *error))completionBlock;
{
	return [self infoForNodeAtResourcePath:node.resourcePath includeChildren:includeChildren completionBlockOnMainQueue:completionBlock];
}

- (id)contentForNode:(U1FileNode*)node progressBlock:(void(^)(long long bytesReceived, long long bytesExpected))progressBlock completionBlock:(void(^)(NSURL *contentURL))completionBlock;
{
	NSString *contentPath = node.contentPath;
	NSString *resourcePath = node.resourcePath;
	
	NSArray *paths = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
	NSString *documentsDirectory = [paths objectAtIndex:0];
	
	NSString *filePath = [documentsDirectory stringByAppendingPathComponent:contentPath];
	NSFileManager *fileManager = [[[NSFileManager alloc] init] autorelease];
	NSString *directory = [filePath stringByDeletingLastPathComponent];
	if (![fileManager fileExistsAtPath:directory isDirectory:NULL])
	{
		[fileManager createDirectoryAtPath:directory withIntermediateDirectories:YES attributes:nil error:NULL];
	}

	NSURL *fileURL = [NSURL fileURLWithPath:filePath];
	if ([fileManager fileExistsAtPath:filePath])
	{
		NSError *error = nil;
		NSDictionary *attributes = [fileManager attributesOfItemAtPath:filePath error:&error];
		
		// if the file doesn't seem to be modified, return from documents directory
		if ([[node size] unsignedLongLongValue] == [attributes fileSize] &&
			[[node modificationDate] compare:[attributes fileModificationDate]] == NSOrderedAscending)
		{
			completionBlock(fileURL);
			return nil;
		}
	}

	NSURL *requestURL = [NSURL URLWithString:[[U1FilesServiceAPIContentRoot stringByAppendingPathComponent:resourcePath] stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]];
	OAToken *token = [[U1AccountManager sharedAccountManager] oauthToken];
	OAConsumer *consumer = [[U1AccountManager sharedAccountManager] oauthConsumer];
	id<OASignatureProviding> signatureProvider = [OAPlaintextSignatureProvider new];
	OAMutableURLRequest *request = [[OAMutableURLRequest alloc] initWithURL:requestURL consumer:consumer token:token realm:nil signatureProvider:signatureProvider];
	[signatureProvider release];
	[request prepare];

	U1HTTPRequestOperationFileDataCollector *fileCollector = [[U1HTTPRequestOperationFileDataCollector alloc] initWithURL:fileURL];
	U1HTTPRequestOperationReportingDataCollector *collector = [[U1HTTPRequestOperationReportingDataCollector alloc] initWithDataCollector:fileCollector];
	[fileCollector release];
	collector.progressBlock = progressBlock;
	U1HTTPRequestOperation *httpOperation = [[U1HTTPRequestOperation alloc] initWithRequest:request dataCollector:collector];
	[httpOperation setCompletionBlock:^(void) {
		[collector release];
		if (completionBlock)
			completionBlock(fileURL);
	}];
	[request release];
	[self.networkQueue addOperation:httpOperation];
	[httpOperation release];
	return nil;
}

- (id)uploadContentData:(id<U1UploadDataProvider>)contentData progressBlock:(void(^)(long long bytesUploaded, long long totalBytes))progressBlock completionBlockOnMainQueue:(void(^)(NSDictionary *updatedNodeInfo, NSError *error))completionBlock;
{
	__block NSString *resourcePath;
	[[U1DataRepository sharedDataRepository] dispatchBlockWithManagedObjectContext:^(NSManagedObjectContext *context) {
		U1Asset *asset = (U1Asset *)[context objectWithID:contentData.assetID];
		NSAssert(asset.fileNode != nil, @"Attempting to upload asset with nil fileNode. Asset is %@ ", asset);
		resourcePath = [asset.fileNode.resourcePath retain];
	}];
	
	NSURL *requestURL = [contentData isMagic] ? [self URLForNode:resourcePath includeChildren:NO] : [self contentURLForNode:resourcePath];
	
	[contentData prepareDataStreamWithBlock:^(NSInputStream *dataStream, NSString *mimeType, NSUInteger length, NSError *error) {
		
		if (error)
		{
			NSLog(@"in uploadContentData, got error '%@' from contentData provider %@", error, contentData);
			dispatch_async(dispatch_get_main_queue(), ^{
				completionBlock(nil, error);
			});
			return;
		}
		
		U1ReportingInputStream *stream = [U1ReportingInputStream inputStreamWithSourceStream:dataStream];
		stream.progressBlock = progressBlock;
		stream.dataLength = length;
		[self execute:@"PUT" toURL:requestURL requestBody:stream contentType:mimeType contentLength:length parseResponseBody:YES completionBlockOnMainQueue:^(id results, NSError *error) {
			completionBlock(results, error);
		}];
	}];
	
	[resourcePath autorelease];
	return nil;
}

- (id)moveNodeAtResourcePath:(NSString*)resourcePath toPath:(NSString*)newPath withCompletionBlockOnMainQueue:(void(^)(NSDictionary *nodeInfo, NSError *error))completionBlock;
{
	NSParameterAssert(resourcePath != nil);
	NSParameterAssert(newPath != nil);
	NSParameterAssert(completionBlock != NULL);
	
	NSURL *nodeURL = [self URLForNode:resourcePath includeChildren:NO];
	NSString *body = [NSString stringWithFormat:@"{\"path\": \"%@\"}", newPath];
	return [self execute:@"PUT" toURL:nodeURL requestBody:body parseResponseBody:YES completionBlockOnMainQueue:^(id results, NSError *error) {
		completionBlock(results, error);
	}];
}

- (id)deleteNodeAtResourcePath:(NSString*)resourcePath completionBlockOnMainQueue:(void(^)(NSError *error))completionBlock;
{
	NSURL *nodeURL = [self URLForNode:resourcePath includeChildren:NO];
	return [self execute:@"DELETE" toURL:nodeURL requestBody:nil parseResponseBody:NO completionBlockOnMainQueue:^(id result, NSError *error) {
		completionBlock(error);
	}];
}

- (id)publishNode:(U1FileNode*)node completionBlockOnMainQueue:(void(^)(U1FileNode *node, NSError *error))completionBlock;
{
	NSString *nodePath = node.resourcePath;
	NSURL *nodeURL = [self URLForNode:(NSString*)nodePath includeChildren:NO];
	return [self execute:@"PUT" toURL:nodeURL requestBody:@"{\"is_public\": true}" parseResponseBody:YES completionBlockOnMainQueue:^(id result, NSError *error) {

		if (!error)
		{
			[node updatePropertiesFromJSONDictionary:result];
		}
		completionBlock(node, error);
	}];
}

- (id)unpublishNode:(U1FileNode*)node completionBlockOnMainQueue:(void(^)(U1FileNode *node, NSError *error))completionBlock;
{
	NSString *nodePath = node.resourcePath;
	NSURL *nodeURL = [self URLForNode:(NSString*)nodePath includeChildren:NO];
	return [self execute:@"PUT" toURL:nodeURL requestBody:@"{\"is_public\": false}" parseResponseBody:YES completionBlockOnMainQueue:^(id result, NSError *error) {

		if (!error)
		{
			[node updatePropertiesFromJSONDictionary:result];
		}
		completionBlock(node, error);
	}];
}

- (id)createFolderAtPath:(NSString*)folderPath completionBlockOnMainQueue:(void(^)(NSDictionary *nodeInfo, NSError *error))completionBlock;
{
	NSParameterAssert(folderPath != nil);
	NSParameterAssert(completionBlock != NULL);
	
	NSURL *nodeURL = [self URLForNode:folderPath includeChildren:NO];
	return [self execute:@"PUT" toURL:nodeURL requestBody:@"{\"kind\": \"directory\"}" parseResponseBody:YES completionBlockOnMainQueue:^(id results, NSError *error) {
		completionBlock(results, error);
	}];
}

- (id)createVolumeAtPath:(NSString*)volumePath completionBlockOnMainQueue:(void(^)(NSDictionary *volumeInfo, NSError *error))completionBlock;
{
	NSParameterAssert(volumePath != nil);
	NSParameterAssert(completionBlock != NULL);
	
	NSURL *volumeURL = [self URLForVolumeWithPath:volumePath];
	return [self execute:@"PUT" toURL:volumeURL requestBody:@"" parseResponseBody:YES completionBlockOnMainQueue:^(id results, NSError *error) {
		completionBlock(results, error);
	}];
}


#pragma mark Private Methods

- (NSURL*)URLForVolumeWithPath:(NSString*)volumePath;
{
	NSParameterAssert(volumePath != nil);
	NSParameterAssert([volumePath hasPrefix:@"/"]);
	
	NSMutableString *urlString = [NSMutableString stringWithString:U1FilesServiceAPIBase];
	[urlString appendString:U1FilesServiceAPIPath];
	[urlString appendString:U1FilesServiceAPIVersion1];
	[urlString appendString:@"/volumes"];
	[urlString appendString:volumePath];
	return [NSURL URLWithString:[urlString stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]];
}

- (NSURL*)URLForNode:(NSString*)nodePath includeChildren:(BOOL)includeChildren;
{
	NSMutableString *urlString = [NSMutableString stringWithString:U1FilesServiceAPIBase];
	[urlString appendString:U1FilesServiceAPIPath];
	[urlString appendString:U1FilesServiceAPIVersion1];
	[urlString appendString:nodePath];
	if (includeChildren)
		[urlString appendString:@"?include_children=true"];
	NSString *finalString = [urlString stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding];
	return [NSURL URLWithString:finalString];
}

- (NSURL *)contentURLForNode:(id)nodePath;
{
	NSMutableString *urlString = [NSMutableString stringWithString:U1FilesServiceAPIContentRoot];
	[urlString appendString:nodePath];
	NSString *finalString = [urlString stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding];
	return [NSURL URLWithString:finalString];
}

- (id)execute:(NSString*)method toURL:(NSURL*)url requestBody:(NSString*)body parseResponseBody:(BOOL)parseResponseBody completionBlockOnMainQueue:(void(^)(id results, NSError *error))completionBlock;
{
	NSData *bodyData = [body dataUsingEncoding:NSUTF8StringEncoding];
	return [self execute:method toURL:url requestBody:bodyData contentType:@"application/json" contentLength:[bodyData length] parseResponseBody:parseResponseBody completionBlockOnMainQueue:completionBlock];
}

- (id)execute:(NSString*)method toURL:(NSURL*)url requestBody:(id)bodyData contentType:(NSString*)contentType contentLength:(NSUInteger)contentLength parseResponseBody:(BOOL)parseResponseBody completionBlockOnMainQueue:(void(^)(id results, NSError *error))completionBlock;
{

	OAToken *token = [[U1AccountManager sharedAccountManager] oauthToken];
	OAConsumer *consumer = [[U1AccountManager sharedAccountManager] oauthConsumer];
	id<OASignatureProviding> signatureProvider = [OAPlaintextSignatureProvider new];
	OAMutableURLRequest *request = [[OAMutableURLRequest alloc] initWithURL:url consumer:consumer token:token realm:nil signatureProvider:signatureProvider];
	[signatureProvider release];
	[request setHTTPMethod:method];
    
	if (bodyData != nil)
	{
		[request setValue:contentType forHTTPHeaderField:@"Content-Type"];
        
		if ([bodyData isKindOfClass:[NSData class]])
		{
			[request setHTTPBody:bodyData];
		}
		else if ([bodyData isKindOfClass:[NSInputStream class]])
		{
			[request setHTTPBodyStream:bodyData];
			[request setValue:[NSString stringWithFormat:@"%u", contentLength] forHTTPHeaderField:@"Content-Length"];
		}
		else
		{
			CLS_LOG(@"unknown bodyData class %@ in request for url %@ callstack:\n%@", bodyData, url, [NSThread callStackSymbols]);
		}
	}
	
	[request prepare];
	
	U1HTTPRequestOperationInMemoryDataCollector *collector = [U1HTTPRequestOperationInMemoryDataCollector new];
	U1HTTPRequestOperation *httpOperation = [[U1HTTPRequestOperation alloc] initWithRequest:request dataCollector:collector];
	[request release];
	
	__block typeof(httpOperation) operationRef = httpOperation;
	[httpOperation setCompletionBlock:^(void) {
			
		id result = nil;
		NSHTTPURLResponse *response = operationRef.response;
		NSError *error = operationRef.error;
		
		if ([response statusCode] == 401)
		{
			dispatch_async(dispatch_get_main_queue(), ^(void) {
				[[U1AccountManager sharedAccountManager] removeCredentials];
			});
			return;
		}
		
		if (!error &&
			![[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(200, 100)] containsIndex:[response statusCode]])
		{
			// TODO: semantic error messages
			error = [NSError errorWithDomain:@"U1FilesServiceErrorDomain" code:[response statusCode] userInfo:nil];
		}
		if (!error && parseResponseBody)
		{
			NSData *data = [collector data];
			result = [NSJSONSerialization JSONObjectWithData:data options:0 error:&error];
			if(result == nil){
				CLS_LOG(@"Error reading JSON %@", error);
			}
		}

		dispatch_async(dispatch_get_main_queue(), ^(void) {

			if (completionBlock)
				completionBlock(result, error);
			
			[collector release];
		});
	}];
	[self.networkQueue addOperation:httpOperation];
	[httpOperation release];
	
	return nil;
}

@end
