//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1FolderViewController.h"

#import <AssetsLibrary/AssetsLibrary.h>
#import <CoreLocation/CoreLocation.h>
#import <MobileCoreServices/UTType.h>
#import <MobileCoreServices/UTCoreTypes.h>

#import "U1Logging.h"

#import "U1Asset.h"
#import "U1AssetManager.h"
#import "U1ByteSizeValueTransformer.h"
#import "U1DataRepository.h"
#import "U1FileNode.h"
#import "U1FilesClient.h"
#import "U1FolderItemCell.h"
#import "U1FolderNode.h"
#import "U1LocalFileInfo.h"
#import "U1AutoUploadsManager.h"
#import "U1Node.h"
#import "UIActionSheet+U1Additions.h"
#import "U1FuzzyDateTransformer.h"


@interface U1FolderViewController () <UINavigationControllerDelegate, UIImagePickerControllerDelegate, NSFetchedResultsControllerDelegate>
@property (retain) IBOutlet UILabel *footerLabel;
@property (copy) NSString *resourcePath;
@property (retain) U1ByteSizeValueTransformer *byteSizeTransformer;
@property (retain) U1FuzzyDateTransformer *fuzzyDateTransformer;
@property (retain) NSIndexPath *pendingIndexPath;
@property (retain) U1FolderNode *node;
@property (retain) NSMutableArray *nodes;
@property (retain) NSFetchedResultsController *resultsController;
- (void)loadInfoForNode;
- (void)deleteNodeAtIndexPath:(NSIndexPath*)indexPath;
- (void)configureCell:(U1FolderItemCell*)cell forNode:(U1Node*)aNode withInfo:(U1LocalFileInfo*)fileInfo;
@end


@implementation U1FolderViewController

@synthesize loadingCell;
@synthesize delegate, filesClient, node, resourcePath;
@synthesize footerLabel, byteSizeTransformer, fuzzyDateTransformer, pendingIndexPath, nodes, resultsController;

- (id)initWithNodeResourcePath:(NSString*)theResourcePath;
{
	self = [super initWithNibName:@"U1FolderViewController" bundle:nil];
	if (!self)
		return nil;
	self.title = [theResourcePath lastPathComponent];
	self.resourcePath = theResourcePath;
	byteSizeTransformer = [U1ByteSizeValueTransformer new];
	fuzzyDateTransformer = [U1FuzzyDateTransformer new];
	filesClient = [[U1FilesClient sharedFilesClient] retain];
	
	UIBarButtonItem *uploadItem = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemAdd target:self action:@selector(uploadImage:)];
	[self.navigationItem setRightBarButtonItem:uploadItem];
	
	[uploadItem release];
	
	return self;
}

- (void)dealloc;
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:U1FilesClientNodeStatusChangedNotification object:nil];
	[filesClient release];
	[loadingCell release];
	[resourcePath release];
	[node release];
	[byteSizeTransformer release];
	[fuzzyDateTransformer release];
	[pendingIndexPath release];
	[nodes release];
	[resultsController release];
	[footerLabel release];
	[super dealloc];
}

- (void)viewDidLoad;
{
	[super viewDidLoad];
	[self refresh];
}

- (void)viewDidUnload;
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:U1FilesClientNodeStatusChangedNotification object:nil];
	[loadingCell release], loadingCell = nil;
	[super viewDidUnload];
}

- (void)viewDidAppear:(BOOL)animated;
{
	[super viewDidAppear:animated];
	[self.tableView deselectRowAtIndexPath:[self.tableView indexPathForSelectedRow] animated:YES];
	
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(nodeStatusChanged:) name:U1FilesClientNodeStatusChangedNotification object:nil];
}

- (void)viewWillDisappear:(BOOL)animated;
{
	[super viewWillDisappear:animated];
	[[NSNotificationCenter defaultCenter] removeObserver:self name:U1FilesClientNodeStatusChangedNotification object:nil];
}

- (void)nodeStatusChanged:(NSNotification*)notification;
{
	U1Node *aNode = [[notification userInfo] objectForKey:@"node"];
	U1LocalFileInfo *fileInfo = [[notification userInfo] objectForKey:@"fileInfo"];
	
	if (aNode.parent != self.node)
		return;
	
	NSIndexPath *indexPath = [self.resultsController indexPathForObject:aNode];
	U1FolderItemCell *cell = (id)[self.tableView cellForRowAtIndexPath:indexPath];
	if (cell)
	{
		[self configureCell:cell forNode:aNode withInfo:fileInfo];
	}
}

- (void)uploadImage:(id)sender;
{
    UIImagePickerController *imagePicker = [[UIImagePickerController alloc] init];
    imagePicker.delegate = self;
    
    if ([CLLocationManager locationServicesEnabled] == YES )
    {
        CLAuthorizationStatus auth = [CLLocationManager authorizationStatus];
        if (auth == kCLAuthorizationStatusAuthorized ||
            auth == kCLAuthorizationStatusNotDetermined)
        {
            [self presentModalViewController:imagePicker animated:YES];
        }
        else
        {
            UIAlertView *alert=[[[UIAlertView alloc] initWithTitle:@"Location Services Off" message:@"U1 Files app must be authorized to use location services to upload from Photos. In Settings App turn on Locatation Services -> U1Files" delegate:nil cancelButtonTitle:@"Ok" otherButtonTitles:nil] autorelease];
            [alert show];
        }
    }
    else
    {
        // present error dialog with directions how to turn it on
        UIAlertView *alert=[[[UIAlertView alloc] initWithTitle:@"Location Services Off" message:@"Location services must be on to upload from Photos. In Settings App turn on Locatation Services" delegate:nil cancelButtonTitle:@"Ok" otherButtonTitles:nil] autorelease];
        [alert show];
    }
}

- (void)imagePickerController:(UIImagePickerController *)picker didFinishPickingMediaWithInfo:(NSDictionary *)info;
{
	// TODO: check for jpg/png
	// TODO: handle videos (stream)
	
	[picker dismissModalViewControllerAnimated:YES];
	
	NSURL *imageURL = [info objectForKey:UIImagePickerControllerReferenceURL];
	U1AutoUploadsManager *autoUploadsManager = [U1AutoUploadsManager sharedAutoUploadsManager];
	ALAssetsLibrary *assetsLibrary = autoUploadsManager.assetsLibrary;
	
	[assetsLibrary assetForURL:imageURL
				   resultBlock:^(ALAsset *asset) {
					   ALAssetRepresentation *representation = [asset defaultRepresentation];
					   CFStringRef uti = (CFStringRef)[representation UTI];
					   NSString *assetType = nil;
					   if (UTTypeConformsTo(uti, kUTTypeImage))
					   {
						   assetType = @"Image";
					   }
					   else if (UTTypeConformsTo(uti, kUTTypeAudiovisualContent))
					   {
						   assetType = @"Video";
					   }
					   else
					   {
						   assetType = @"File"; // XXX: Extend to something better when we bring in external files
					   }
					   NSDate *assetDate = [asset valueForProperty:ALAssetPropertyDate];

					   NSDateFormatter *dateFormatter = [[NSDateFormatter alloc] init];
					   [dateFormatter setDateStyle:NSDateFormatterMediumStyle];
					   [dateFormatter setTimeStyle:NSDateFormatterMediumStyle];
					   
					   NSString *ext = [(id)UTTypeCopyPreferredTagWithClass(uti, kUTTagClassFilenameExtension) autorelease];
					   NSString *filename = [NSString stringWithFormat:@"%@ %@.%@", assetType, [dateFormatter stringFromDate:assetDate], ext];
					   [dateFormatter release];
					   
					   dispatch_async(dispatch_get_main_queue(), ^(void) {
						   
						   __block U1Asset *u1asset = nil;
						   [[U1DataRepository sharedDataRepository] dispatchBlockWithMainContext:^(NSManagedObjectContext *mainContext){
							   u1asset = [[U1AssetManager sharedAssetManager] createU1AssetWithAsset:asset
																			  group:nil
																		  URLString:[imageURL absoluteString]
																		   fileName:filename
																			context:mainContext];
							   [[U1AssetManager sharedAssetManager] findOrCreateMatchingFileNodeForU1Asset:u1asset inFolderNode:self.node context:mainContext];
						   }];
						   
						   if (u1asset.generation == nil) // Prevent duplicate uploading to the same folder
						   {
							   [self.filesClient uploadAssetWithID:[u1asset objectID]
												   withPriority:NSOperationQueuePriorityVeryHigh
												  progressBlock:^(long long bytesUploaded, long long totalBytes) {}
									 completionBlockOnMainQueue:^(U1FileNode *updatedNode, NSError *error) {
										 
										 if (error){
											 CLS_LOG(@"Error uploading asset, attempting to delete related filenode");
											 // TODO: remove direct access to data repository
											 [[U1DataRepository sharedDataRepository] dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
												 if (u1asset.fileNode == nil) {
													 CLS_LOG(@"u1asset fileNode is nil");
												 }else{
													 [context deleteObject:u1asset.fileNode];
												 }
											 }];
										 }
										 else
										 {
											 u1asset.generation = updatedNode.generation;
										 }
									 }];
						   }
						   
						   NSError *err;
						   BOOL success = [[U1DataRepository sharedDataRepository] save:&err];
						   if(!success){
							   CLS_LOG(@"Error saving u1asset creation: %@", err);
						   }
					   });

				   } failureBlock:^(NSError *error) {
					   NSLog(@"Couldn't fetch asset from library: %@", error);
				   }];
	
}

- (void)configureCell:(U1FolderItemCell*)cell forNode:(U1Node*)aNode withInfo:(U1LocalFileInfo*)fileInfo;
{
	cell.nameLabel.text = [aNode.resourcePath lastPathComponent];
	
	BOOL hasChildren = [[aNode hasChildren] boolValue];
	[cell setAccessoryType:(hasChildren ? UITableViewCellAccessoryDisclosureIndicator : UITableViewCellAccessoryNone)];
	
	if (aNode.generation == nil)
	{
		[cell.uploadIndicator setHidden:NO];
		[cell.uploadIndicator startAnimating];
	}
	else
	{
		[cell.uploadIndicator setHidden:YES];
		[cell.uploadIndicator stopAnimating];
	}
	
	[cell.imageView setImage:nil];
	if ([aNode isFolder])
	{
		[cell.imageView setImage:[UIImage imageNamed:@"ic_folder"]];
		NSString *fuzzyDate = [self.fuzzyDateTransformer transformedValue:aNode.modificationDate];
		[cell.detailLabel setText:[NSString stringWithFormat:@"modified %@", fuzzyDate]];
	}
	else
	{
		if(fileInfo){
			[cell.uploadProgressView setProgress:fileInfo.uploadPercentage];
			[cell.uploadProgressView setHidden:!fileInfo.isUploading];
			[cell.detailLabel setHidden:fileInfo.isUploading];
		}else{
			[cell.uploadProgressView setHidden:YES];
			[cell.detailLabel setHidden:NO];
		}

		NSString *contentPath = aNode.contentPath;
		NSURL *contentURL = [NSURL fileURLWithPath:contentPath];
		UIDocumentInteractionController *docController = [UIDocumentInteractionController interactionControllerWithURL:contentURL];
		[cell.imageView setImage:[docController.icons lastObject]];
		NSNumber *size = [aNode valueForKey:@"size"];
		NSString *sizeString = [self.byteSizeTransformer transformedValue:size];
		NSString *detailString = sizeString;
		NSString *fuzzyDate = [self.fuzzyDateTransformer transformedValue:aNode.modificationDate];
		if (aNode.modificationDate)
			detailString = [sizeString stringByAppendingFormat:@", modified %@", fuzzyDate];
		[cell.detailLabel setText:detailString];
	}
}


#pragma mark UITableViewDataSource Methods

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section;
{
	if (!self.resultsController)
		return 1;
	
	return [[[self.resultsController sections] objectAtIndex:section] numberOfObjects];
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath;
{
	if (!self.resultsController)
	{
		return self.loadingCell;
	}
	
	U1FolderItemCell *cell = [U1FolderItemCell cellForTableView:tableView];
	
	U1Node *childNode = [self.resultsController objectAtIndexPath:indexPath];
	

	[self configureCell:cell forNode:childNode withInfo:nil];
	
	return cell;
}


#pragma mark UITableViewDelegate Methods

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath;
{
	U1Node *childNode = [self.resultsController objectAtIndexPath:indexPath];
	[self.delegate folderController:self didSelectNode:childNode];
}

- (void)tableView:(UITableView *)tableView commitEditingStyle:(UITableViewCellEditingStyle)editingStyle forRowAtIndexPath:(NSIndexPath *)indexPath;
{
	if (editingStyle != UITableViewCellEditingStyleDelete)
		return;
	
	[self deleteNodeAtIndexPath:indexPath];
}


- (void)refresh;
{
	[self loadInfoForNode];
}


#pragma mark Private Methods

- (void)loadInfoForNode;
{
	NSParameterAssert(self.resourcePath != nil);

	[self.filesClient fetchFolderWithResourcePath:self.resourcePath resultsBlock:^(U1Node *theNode, NSFetchedResultsController *childrenResultsController, U1DataSourceType resultsDataSourceType) {
		
		self.node = (id)theNode;
		if (childrenResultsController != self.resultsController)
		{
			self.resultsController = childrenResultsController;
			[self.resultsController setDelegate:self];
			[self.tableView reloadData];
		}

		if (resultsDataSourceType == U1DataSourceRemote)
			[self stopLoading];
		
	} errorBlock:^(NSError *error) {
		NSLog(@"Couldn't fetch remote folder: %@", error);
		[self stopLoading];
	}];
}

- (void)deleteNodeAtIndexPath:(NSIndexPath*)indexPath;
{
	U1Node *childNode = [self.resultsController objectAtIndexPath:indexPath];
	[self.filesClient deleteNode:childNode completionBlock:^(NSError *error) {
		if (error)
		{
			// TODO:
			CLS_LOG(@"Error deleting node: %@", error);
			return;
		}
	}];
}


#pragma mark NSFetchedResultsControllerDelegate Methods


- (void)controllerWillChangeContent:(NSFetchedResultsController *)controller;
{
	[self.tableView beginUpdates];
}


- (void)controller:(NSFetchedResultsController *)controller didChangeSection:(id <NSFetchedResultsSectionInfo>)sectionInfo
		   atIndex:(NSUInteger)sectionIndex forChangeType:(NSFetchedResultsChangeType)type;
{
	switch (type)
	{
		case NSFetchedResultsChangeInsert:
			[self.tableView insertSections:[NSIndexSet indexSetWithIndex:sectionIndex]
						  withRowAnimation:UITableViewRowAnimationFade];
			break;
			
		case NSFetchedResultsChangeDelete:
			[self.tableView deleteSections:[NSIndexSet indexSetWithIndex:sectionIndex]
						  withRowAnimation:UITableViewRowAnimationFade];
			break;
	}
}

- (void)controller:(NSFetchedResultsController *)controller didChangeObject:(id)anObject
	   atIndexPath:(NSIndexPath *)indexPath forChangeType:(NSFetchedResultsChangeType)type
	  newIndexPath:(NSIndexPath *)newIndexPath;
{
	
	UITableView *tableView = self.tableView;
	
	switch (type)
	{
			
		case NSFetchedResultsChangeInsert:
			[tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:newIndexPath]
							 withRowAnimation:UITableViewRowAnimationMiddle];
			break;
			
		case NSFetchedResultsChangeDelete:
			[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
							 withRowAnimation:UITableViewRowAnimationTop];
			break;
			
		case NSFetchedResultsChangeUpdate:
			// TODO: 
			//            [self configureCell:[tableView cellForRowAtIndexPath:indexPath]
			//					atIndexPath:indexPath];
			[self.tableView reloadRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
								  withRowAnimation:UITableViewRowAnimationNone];
			break;
			
		case NSFetchedResultsChangeMove:
			[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			[tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:newIndexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			break;
	}
}

- (void)controllerDidChangeContent:(NSFetchedResultsController *)controller;
{
	[self.tableView endUpdates];
	
	int numItems = [[[self.resultsController sections] objectAtIndex:0] numberOfObjects];
	NSNumber *folderRawSize = [[self.resultsController fetchedObjects] valueForKeyPath:@"@sum.size"];
	NSString *folderSize = [self.byteSizeTransformer transformedValue:folderRawSize];
	[self.footerLabel setText:[NSString stringWithFormat:@"%d items, %@", numItems, folderSize]];
}

@end
