//
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1HTTPRequestOperation.h"

#import "UONetworkStatusCoordinator.h"


@interface U1HTTPRequestOperation ()
@property (getter=isExecuting) BOOL executing;
@property (getter=isFinished) BOOL finished;
@property (assign) id<U1HTTPRequestOperationDataCollector> dataCollector;
@property (retain) NSURLConnection *connection;
@property (readwrite, copy) NSHTTPURLResponse *response;
@property (readwrite, retain) NSError *error;
- (void)finishExecuting;
- (BOOL)checkForCancellation;
@end


@implementation U1HTTPRequestOperation

@synthesize credential;
@synthesize executing, finished;
@synthesize dataCollector, connection, response, error;

- (id)initWithRequest:(NSURLRequest*)request dataCollector:(id<U1HTTPRequestOperationDataCollector>)theDataCollector;
{
	if (!(self = [super init]))
		return nil;
	
	connection = [[NSURLConnection alloc] initWithRequest:request delegate:self startImmediately:NO];
	[connection scheduleInRunLoop:[NSRunLoop mainRunLoop] forMode:NSDefaultRunLoopMode];
	dataCollector = theDataCollector;
	
	return self;
}

- (void)dealloc;
{
	[credential release];
	dataCollector = nil;
	[connection cancel];
	[connection release];
	[response release];
	[error release];
	[super dealloc];
}


#pragma mark Private Methods

- (void)finishExecuting;
{
	[self willChangeValueForKey:@"isExecuting"];
	[self willChangeValueForKey:@"isFinished"];
	self.finished = YES;
	self.executing = NO;
	[self didChangeValueForKey:@"isFinished"];
	[self didChangeValueForKey:@"isExecuting"];
	[UONetworkStatusCoordinator removeNetworkActivity];
}

- (BOOL)checkForCancellation;
{
	if ([self isCancelled])
	{
		[self.connection cancel];
		[self finishExecuting];
		return YES;
	}
	return NO;
}


#pragma mark NSOperation Methods

- (BOOL)isConcurrent;
{
	return YES;
}

- (void)start;
{
	if ([self checkForCancellation])
	{
		return;
	}
	
	[self.connection start];
	[UONetworkStatusCoordinator addNetworkActivity];
	
	[self willChangeValueForKey:@"isExecuting"];
	self.executing = YES;
	[self didChangeValueForKey:@"isExecuting"];
}

- (void)cancel;
{
	[self.connection cancel];
	[super cancel];
}


#pragma mark NSURLConnectionDelegate Methods

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)theResponse;
{
	NSParameterAssert([theResponse isKindOfClass:[NSHTTPURLResponse class]]);
	
	self.response = (NSHTTPURLResponse*)theResponse;

	if ([self checkForCancellation])
		return;
	
	if ([self.dataCollector respondsToSelector:@selector(httpRequestOperation:willExpectDataOfLength:)])
		[self.dataCollector httpRequestOperation:self willExpectDataOfLength:[self.response expectedContentLength]];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)dataChunk;
{
	if ([self checkForCancellation])
		return;
	
	if ([self.dataCollector respondsToSelector:@selector(httpRequestOperation:didReceiveData:)])
		[self.dataCollector httpRequestOperation:self didReceiveData:dataChunk];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection;
{
	if ([self.dataCollector respondsToSelector:@selector(httpRequestOperationDidFinishLoadingData:)])
		[self.dataCollector httpRequestOperationDidFinishLoadingData:self];
	[self finishExecuting];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)anError;
{
	self.error = anError;
	if ([self.dataCollector respondsToSelector:@selector(httpRequestOperation:didFailWithError:)])
		[self.dataCollector httpRequestOperation:self didFailWithError:self.error];
	[self finishExecuting];
}

- (void)connection:(NSURLConnection *)connection didReceiveAuthenticationChallenge:(NSURLAuthenticationChallenge *)challenge;
{
	if (self.credential && [challenge previousFailureCount] == 0)
		[[challenge sender] useCredential:self.credential forAuthenticationChallenge:challenge];
	else
		[[challenge sender] continueWithoutCredentialForAuthenticationChallenge:challenge];
}

@end


// TODO: move this into its own file

@interface U1HTTPRequestOperationInMemoryDataCollector ()
{
	NSMutableData *collectedData;
}
@end


@implementation U1HTTPRequestOperationInMemoryDataCollector

@synthesize data=collectedData;

- (id)init;
{
	if (!(self = [super init]))
		return nil;

	return self;
}

- (void)dealloc;
{
	[collectedData release];
	[super dealloc];
}


#pragma mark U1HTTPRequestOperationDataCollector Methods

- (void)httpRequestOperation:(U1HTTPRequestOperation*)httpOperation willExpectDataOfLength:(long long)length;
{
	if (length == NSURLResponseUnknownLength)
		length = 0;

	[collectedData release], collectedData = nil;
	collectedData = [[NSMutableData alloc] initWithCapacity:length];
}

- (void)httpRequestOperation:(U1HTTPRequestOperation *)httpOperation didReceiveData:(NSData*)data;
{
	[collectedData appendData:data];
}

- (void)httpRequestOperationDidFinishLoadingData:(U1HTTPRequestOperation*)httpOperation;
{
	
}

- (void)httpRequestOperation:(U1HTTPRequestOperation*)httpOperation didFailWithError:(NSError*)error;
{
	
}

@end



@interface U1HTTPRequestOperationFileDataCollector ()
@property (retain) NSOutputStream *fileStream;
@end


@implementation U1HTTPRequestOperationFileDataCollector

@synthesize fileURL;
@synthesize fileStream;

- (id)initWithURL:(NSURL*)theURL;
{
	if (!(self = [super init]))
		return nil;
	fileURL = [theURL copy];
	fileStream = [[NSOutputStream alloc] initWithURL:fileURL append:NO];
	[fileStream open];
	return self;
}

- (void)dealloc;
{
	[fileURL release];
	[fileStream release];
	[super dealloc];
}


#pragma mark U1HTTPRequestOperationDataCollector Methods

- (void)httpRequestOperation:(U1HTTPRequestOperation*)httpOperation willExpectDataOfLength:(long long)length;
{
}

- (void)httpRequestOperation:(U1HTTPRequestOperation *)httpOperation didReceiveData:(NSData*)data;
{
	[fileStream write:[data bytes] maxLength:[data length]];
}

- (void)httpRequestOperationDidFinishLoadingData:(U1HTTPRequestOperation*)httpOperation;
{
	[fileStream close];
}

- (void)httpRequestOperation:(U1HTTPRequestOperation*)httpOperation didFailWithError:(NSError*)error;
{
	[fileStream close];
}

@end


@interface U1HTTPRequestOperationReportingDataCollector ()
@property (retain) id<U1HTTPRequestOperationDataCollector> nextCollector;
@property long long expectedLength;
@property long long currentLength;
@end

@implementation U1HTTPRequestOperationReportingDataCollector

@synthesize progressBlock;
@synthesize nextCollector, expectedLength, currentLength;

- (id)initWithDataCollector:(id<U1HTTPRequestOperationDataCollector>)theNextCollector;
{
	if (!(self = [super init]))
		return nil;
	
	nextCollector = [theNextCollector retain];
	
	return self;
}

- (void)dealloc;
{
	[nextCollector release];
	[progressBlock release];
	[super dealloc];
}

#pragma mark U1HTTPRequestOperationDataCollector Methods

- (void)httpRequestOperation:(U1HTTPRequestOperation*)httpOperation willExpectDataOfLength:(long long)length;
{
	[self.nextCollector httpRequestOperation:httpOperation willExpectDataOfLength:length];
	
	self.expectedLength = length;
	self.currentLength = 0;
	if (self.progressBlock)
		self.progressBlock(self.currentLength, self.expectedLength);
}

- (void)httpRequestOperation:(U1HTTPRequestOperation *)httpOperation didReceiveData:(NSData*)data;
{
	[self.nextCollector httpRequestOperation:httpOperation didReceiveData:data];
	
	self.currentLength += [data length];
	if (self.progressBlock)
		self.progressBlock(self.currentLength, self.expectedLength);
}

- (void)httpRequestOperationDidFinishLoadingData:(U1HTTPRequestOperation*)httpOperation;
{
	[self.nextCollector httpRequestOperationDidFinishLoadingData:httpOperation];
}

- (void)httpRequestOperation:(U1HTTPRequestOperation*)httpOperation didFailWithError:(NSError*)error;
{
	[self.nextCollector httpRequestOperation:httpOperation didFailWithError:error];
}

@end
