//
//  Copyright 2013 Canonical Ltd.
//
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


#import "U1Hasher.h"

#import <CommonCrypto/CommonDigest.h>

NSString * const U1ContentHashKey = @"contentHash";
NSString * const U1MagicHashKey = @"magicHash";

@implementation U1Hasher



+ (NSData *)contentHashForAssetData:(NSData *)imageData length:(uint32_t)length{
	
	unsigned char* hashDigest = (unsigned char *)malloc(CC_SHA1_DIGEST_LENGTH);
	
	CC_SHA1([imageData bytes], length, hashDigest);
	return [NSData dataWithBytesNoCopy:hashDigest length:CC_SHA1_DIGEST_LENGTH];
	
}

+ (NSData *)magicHashForAssetData:(NSData *)imageData length:(uint32_t)length{
	
	unsigned char* magicDigest = (unsigned char*)malloc(CC_SHA1_DIGEST_LENGTH);
	
	CC_SHA1_CTX magicCtx;
	CC_SHA1_Init(&magicCtx);
	CC_SHA1_Update(&magicCtx, "Ubuntu One", 10);
	CC_SHA1_Update(&magicCtx, [imageData bytes], length);
	CC_SHA1_Final(magicDigest, &magicCtx);
	
	return [NSData dataWithBytesNoCopy:magicDigest length:CC_SHA1_DIGEST_LENGTH];
}

- (id)init{
	if (!(self = [super init])){
		return nil;
	}
	
	self.assetsLibrary = [[[ALAssetsLibrary alloc] init] autorelease];
	sync_semaphore = dispatch_semaphore_create(0);
	return self;
}

- (void)dealloc{
	dispatch_release(sync_semaphore);
	[super dealloc];
}

// because assetForURL can be asynchronous if it needs to ask user permission,
// we use a semaphore to wait for the hashing.
// but, because it may be synchronous if it doesn't need permission, we wrap it
// in an async block so the semaphore won't deadlock.
- (NSDictionary *)synchronouslyGetHashesForAssetURL:(NSURL *)assetURL{
	__block NSMutableDictionary *hashDict = nil;
	
	dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_HIGH, 0), ^{
		[self.assetsLibrary assetForURL:assetURL
							resultBlock:^(ALAsset *asset) {
								
            ALAssetRepresentation *representation = [asset defaultRepresentation];
			uint8_t *buffer = (uint8_t*)malloc(representation.size);
			NSUInteger length = [representation getBytes:buffer fromOffset:0 length:representation.size error:nil];
			NSData *imageData = [NSData dataWithBytesNoCopy:buffer length:length freeWhenDone:YES];
			
			hashDict = [NSMutableDictionary dictionaryWithObjectsAndKeys:
						[U1Hasher contentHashForAssetData:imageData length:length], U1ContentHashKey,
						[U1Hasher magicHashForAssetData:imageData length:length], U1MagicHashKey, nil];
			[hashDict retain];
								
			dispatch_semaphore_signal(sync_semaphore);
			
			}failureBlock:^(NSError *error) {
				dispatch_semaphore_signal(sync_semaphore);
			}];
	});
	
	dispatch_semaphore_wait(sync_semaphore, DISPATCH_TIME_FOREVER);
	
	return [hashDict autorelease];
}
@end
