//
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1LoginController.h"

#import "OAuthConsumer.h"
#import "NSString+Extras.h"

#import "U1AccountManager.h"
#import "UOHTTPFetchOperation.h"
#import "UOJSONFetchOperation.h"
#import "U1SSOCredentialsViewController.h"
#import "U1SSOLandingPageViewController.h"
#import "U1SSOSignUpViewController.h"
#import "U1SingleSignOnProvider.h"
#import "U1SSOEmailVerificationViewController.h"
#import "U1SSOAuthenticateResponse.h"
#import "U1HUDAlert.h"
#import "SVProgressHUD.h"


static NSString *const UOSingleSignOnBaseURL = @"https://login.ubuntu.com/api/1.0/";

static NSString *UOSSOEncodedStringFromParametersDictionary(NSDictionary *arguments);


@interface U1LoginController () <UINavigationControllerDelegate>
@property (retain) IBOutlet UINavigationController *loginViewController;
@property (retain) IBOutlet U1SSOCredentialsViewController *rootController;
@property (retain) IBOutlet U1SSOLandingPageViewController *landingPageController;
@property (retain) U1SingleSignOnProvider *ssoProvider;
@property (retain) NSOperationQueue *operationQueue;
- (void)loginWithUsername:(NSString*)username password:(NSString*)password;
- (void)handleError:(NSError*)error;
@end


@implementation U1LoginController

@synthesize delegate, loginViewController, rootController, landingPageController, ssoProvider, operationQueue;

- (id)init;
{
	if (!(self = [super init]))
		return nil;

	__block id this = self;

	ssoProvider = [[U1SingleSignOnProvider alloc] initWithNetworkQueue:nil];
	[[UINib nibWithNibName:@"U1LoginController" bundle:nil] instantiateWithOwner:self options:nil];
	[self.loginViewController setNavigationBarHidden:YES animated:NO];
	[self.loginViewController setDelegate:self];
	
	landingPageController = [[U1SSOLandingPageViewController alloc] initWithNibName:nil bundle:nil];
	landingPageController.signInAction = ^(id sender) {
		[this signIn:sender];
	};
	landingPageController.signUpAction = ^(id sender) {
		[this signUp:sender];
	};
	
	NSArray *controllers = [NSArray arrayWithObject:landingPageController];
	[self.loginViewController setViewControllers:controllers animated:NO];

	rootController = [[U1SSOCredentialsViewController alloc] initWithNibName:@"U1SSOCredentialsViewController" bundle:nil];
	rootController.navigationItem.titleView = [[[UIImageView alloc] initWithImage:[UIImage imageNamed:@"ios_form_logo"]] autorelease];
	
	rootController.loginAction = ^ (NSString *username, NSString *password)
	{
		[this loginWithUsername:username password:password];
	};
	
	operationQueue = [[NSOperationQueue alloc] init];
	
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationActivated:) name:UIApplicationDidBecomeActiveNotification object:nil];
	
	return self;
}

- (void)dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:UIApplicationDidBecomeActiveNotification object:nil];
	delegate = nil;
	[loginViewController release];
	[rootController release];
	[landingPageController release];
	[ssoProvider release];
	[operationQueue release];
    [super dealloc];
}

- (void)applicationActivated:(NSNotification*)notification;
{
//	if (![[U1AccountManager sharedAccountManager] isVerified])
//	{
//		[self continueSignUp:self];
//	}
}

- (IBAction)signIn:(id)sender;
{
	[self.loginViewController pushViewController:rootController animated:YES];
}

- (IBAction)signUp:(id)sender;
{
	U1SSOSignUpViewController *signUpController = [[[U1SSOSignUpViewController alloc] initWithNibName:nil bundle:nil] autorelease];
	signUpController.navigationItem.titleView = [[[UIImageView alloc] initWithImage:[UIImage imageNamed:@"ios_form_logo"]] autorelease];
	signUpController.ssoProvider = self.ssoProvider;
	__block U1SSOSignUpViewController *signUpControllerRef = signUpController;
	signUpController.signUpAction = ^(NSString *fullName, NSString *email, NSString *password, NSString *captchaId, NSString *captchaResponse) {
		
		[SVProgressHUD showWithStatus:@"Signing Up" maskType:SVProgressHUDMaskTypeBlack networkIndicator:YES];
		
		[self.ssoProvider registerAccountForEmail:email withPassword:password displayName:fullName captchaId:captchaId captchaSolution:captchaResponse completionHandlerOnMainQueue:^(U1SSOResponse *response, NSError *error) {
			
			if (error ||
			    [[response status] isEqualToString:@"error"])
			{
				[SVProgressHUD dismissWithError:@"Sign Up Failed"];
				[signUpControllerRef presentErrors:[response errors]];
			}
			else
			{
				NSString *tokenName = [NSString stringWithFormat:@"Ubuntu One @ %@", [[UIDevice currentDevice] name]];
				[self.ssoProvider authenticateAccountForEmail:email withPassword:password appToken:tokenName completionHandlerOnMainQueue:^(U1SSOAuthenticateResponse *response, NSError *error) {
                    
					[SVProgressHUD dismiss];
                    
					// save temporary tokens
					OAConsumer *consumer = [[[OAConsumer alloc] initWithKey:response.consumerKey secret:response.consumerSecret] autorelease];
					OAToken *token = [[[OAToken alloc] initWithKey:response.token secret:response.tokenSecret] autorelease];
                    
					[[U1AccountManager sharedAccountManager] storeConsumerKey:consumer.key consumerSecret:consumer.secret token:token.key tokenSecret:token.secret];
                    
					// TODO:indicate the account is {un}verified
                    
					// push the "verify your email screen"
					NSMutableArray *controllers = [[self.loginViewController viewControllers] mutableCopy];
					[controllers removeLastObject];
					U1SSOEmailVerificationViewController *verifyController = [[[U1SSOEmailVerificationViewController alloc] initWithNibName:nil bundle:nil] autorelease];
					verifyController.navigationItem.titleView = [[[UIImageView alloc] initWithImage:[UIImage imageNamed:@"ios_form_logo"]] autorelease];
					[controllers addObject:verifyController];
					[self.loginViewController setViewControllers:controllers animated:YES];
                    
					}];
			}
			
			[SVProgressHUD dismiss];
            
		}];
	};
	[self.loginViewController pushViewController:signUpController animated:YES];
}

- (IBAction)continueSignUp:(id)sender;
{
	U1AccountManager *accountManager = [U1AccountManager sharedAccountManager];
	// TODO: show a loading screen
	
	// TODO: get oauth tokens
	OAConsumer *consumer = [accountManager oauthConsumer];
	OAToken *token = [accountManager oauthToken];
	
	// TODO: attempt request to get login details
	OAPlaintextSignatureProvider *signatureProvider = [[[OAPlaintextSignatureProvider alloc] init] autorelease];
	NSDictionary *params = [NSDictionary dictionaryWithObjectsAndKeys:
							@"me", @"ws.op",
							nil];
	NSString * paramsString = UOSSOEncodedStringFromParametersDictionary(params);
	
	NSURL *meURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@accounts?%@", UOSingleSignOnBaseURL, paramsString]];
	OAMutableURLRequest *oauthRequest = [[[OAMutableURLRequest alloc] initWithURL:meURL consumer:consumer token:token realm:nil signatureProvider:signatureProvider] autorelease];
	[oauthRequest prepare];
	UOJSONFetchOperation *tokenOperation = [UOJSONFetchOperation jsonOperationWithRequest:oauthRequest action:^(NSObject *responseBody, NSError *error) {
		
		// 3. Get the phone creds using the new token
		BOOL verified = NO;
		
		NSString *primaryEmail = nil;
		if (error)
		{
			// TODO: show a message about failure?
		}
		else
		{
			NSDictionary *result = (id)responseBody;
			primaryEmail = [result objectForKey:@"preferred_email"];
			verified = primaryEmail && !(primaryEmail == (id)kCFNull);
		}
		
		[accountManager setVerified:verified];
		
		if (verified)
		{
			
			NSString *encodedUsername = [primaryEmail urlParameterEncodedString];
			NSURL *u1TokenURL = [NSURL URLWithString:[NSString stringWithFormat:@"https://one.ubuntu.com/oauth/sso-finished-so-get-tokens/%@", encodedUsername]];
			
			OAMutableURLRequest *oauthRequest = [[[OAMutableURLRequest alloc] initWithURL:u1TokenURL consumer:consumer token:token realm:nil signatureProvider:signatureProvider] autorelease];
			[oauthRequest prepare];
			UOHTTPFetchOperation *tokenOperation = [UOHTTPFetchOperation httpOperationWithRequest:oauthRequest action:^(NSString *responseBody, NSError *error) {
				
				if (error)
				{
					[self handleError:error];
					return;
				}
				
				[self.delegate loginControllerDidLogin];
				[self.loginViewController.view setUserInteractionEnabled:YES];
				
			}];
			[self.operationQueue addOperation:tokenOperation];
		}
		else
		{
			U1SSOEmailVerificationViewController *verifyController = [[[U1SSOEmailVerificationViewController alloc] initWithNibName:nil bundle:nil] autorelease];
			verifyController.navigationItem.titleView = [[[UIImageView alloc] initWithImage:[UIImage imageNamed:@"ios_form_logo"]] autorelease];
			NSArray *controllers = [NSArray arrayWithObjects:
									self.landingPageController,
									verifyController,
									nil];
			[self.loginViewController setViewControllers:controllers animated:YES];
			UIViewController *modalPresenter = [[[UIApplication sharedApplication] keyWindow] rootViewController];
			[modalPresenter presentModalViewController:self.loginViewController animated:YES];
		}
		
	}];
	[self.operationQueue addOperation:tokenOperation];

	// if success, verified and completed
	
	// if fail, show waiting for verification page
	
}

- (void)loginWithUsername:(NSString*)username password:(NSString*)password;
{
	// TODO: lots of this should use U1SingleSignOnProvider now...
	
	// - show login spinner
	[self.loginViewController.view setUserInteractionEnabled:NO];
	
	// 1. get/create U1 oauth token
	NSString *tokenName = [NSString stringWithFormat:@"Ubuntu One @ %@", [[UIDevice currentDevice] name]];
	NSDictionary *params = [NSDictionary dictionaryWithObjectsAndKeys:
							@"authenticate", @"ws.op",
							tokenName, @"token_name",
							nil];
	NSString * paramsString = UOSSOEncodedStringFromParametersDictionary(params);
	
	NSURL *url = [NSURL URLWithString:[NSString stringWithFormat:@"%@authentications?%@", UOSingleSignOnBaseURL, paramsString]];
	NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:url];
	[request setHTTPMethod:@"GET"];
	
	NSURLCredential *credential = [NSURLCredential credentialWithUser:username password:password persistence:NSURLCredentialPersistenceNone];
	UOJSONFetchOperation *jsonOperation = [UOJSONFetchOperation jsonOperationWithRequest:request action:^(NSObject *jsonResult, NSError *error) {
		
		// 2. import the new token into Ubuntu One
		NSDictionary *oauthResponse = (id)jsonResult;
		
		if (error)
		{
			[self handleError:error];
			return;
		}
		
		NSString *encodedUsername = [username urlParameterEncodedString];
		NSURL *u1TokenURL = [NSURL URLWithString:[NSString stringWithFormat:@"https://one.ubuntu.com/oauth/sso-finished-so-get-tokens/%@", encodedUsername]];
		
		OAConsumer *consumer = [[[OAConsumer alloc] initWithKey:[oauthResponse objectForKey:@"consumer_key"] secret:[oauthResponse objectForKey:@"consumer_secret"]] autorelease];
		OAToken *token = [[[OAToken alloc] initWithKey:[oauthResponse objectForKey:@"token"] secret:[oauthResponse objectForKey:@"token_secret"]] autorelease];

		[[U1AccountManager sharedAccountManager] storeConsumerKey:consumer.key consumerSecret:consumer.secret token:token.key tokenSecret:token.secret];
		
		OAPlaintextSignatureProvider *signatureProvider = [[[OAPlaintextSignatureProvider alloc] init] autorelease];
		OAMutableURLRequest *oauthRequest = [[[OAMutableURLRequest alloc] initWithURL:u1TokenURL consumer:consumer token:token realm:nil signatureProvider:signatureProvider] autorelease];
		[oauthRequest prepare];
		UOHTTPFetchOperation *tokenOperation = [UOHTTPFetchOperation httpOperationWithRequest:oauthRequest action:^(NSString *responseBody, NSError *error) {
			
			// 3. Get the phone creds using the new token
			
			if (error)
			{
				[self handleError:error];
				return;
			}
			
			[self.delegate loginControllerDidLogin];
			[self.loginViewController.view setUserInteractionEnabled:YES];
		
		}];
		[self.operationQueue addOperation:tokenOperation];
		
	}];
	jsonOperation.credential = credential;
	[self.operationQueue addOperation:jsonOperation];
}

- (void)handleError:(NSError*)error;
{
	NSString *errorMessage = [error localizedDescription];
	if ([error code] == 401)
		errorMessage = NSLocalizedString(@"Incorrect username or password.", nil);
	else if (errorMessage == nil)
		errorMessage = NSLocalizedString(@"An error occured trying to sign in, please try again.", @"Login Error Message");

	UIAlertView *alert = [[UIAlertView alloc] initWithTitle:NSLocalizedString(@"Ubuntu One", nil) message:errorMessage delegate:self cancelButtonTitle:@"Dismiss" otherButtonTitles:nil];
	[alert show];
	[alert release];
	[self.loginViewController.view setUserInteractionEnabled:YES];
	U1SSOCredentialsViewController *loginController = (id)[(UINavigationController*)[self loginViewController] topViewController];
	[loginController reset];
}

- (void)navigationController:(UINavigationController *)navigationController willShowViewController:(UIViewController *)viewController animated:(BOOL)animated;
{
	BOOL show = [navigationController.viewControllers indexOfObjectIdenticalTo:viewController] == 0;
	[navigationController setNavigationBarHidden:show animated:YES];
}

@end


static NSString *UOSSOEncodedStringFromParametersDictionary(NSDictionary *arguments)
{
	NSMutableArray *encodedArguments = [NSMutableArray array];
	[arguments enumerateKeysAndObjectsUsingBlock:^(id key, id value, BOOL *stop) {
		if (![key isEqualToString:@"ws.op"]) {
            value = [NSString stringWithFormat:@"\"%@\"", value];
        }
		value = [value urlParameterEncodedString];
		NSString *argumentPair = [key stringByAppendingFormat:@"=%@", value];
		[encodedArguments addObject:argumentPair];
	}];
	
	return [encodedArguments componentsJoinedByString:@"&"];
}
