//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1SettingsViewController.h"

#import "U1Logging.h"

#import <CoreLocation/CoreLocation.h>
#import <QuartzCore/QuartzCore.h>

#import "U1AccountManager.h"
#import "U1ByteSizeValueTransformer.h"
#import "U1DataRepository.h"
#import "U1FilesClient.h"


enum U1SettingsViewSection {
	U1SettingsViewSectionAccount,
	U1SettingsViewSectionCameraRoll,
};
static const NSUInteger U1SettingsViewSectionCount = 2;


@interface U1SettingsViewController () <UIActionSheetDelegate>
@property (readonly, nonatomic, retain) UITableViewCell *accountLoadingCell;
@property (retain) U1FilesClient *filesClient;
@property (retain) NSDictionary *accountInfo;
- (UITableViewCell *)preparedAutoUploadCell;
- (UITableViewCell *)preparedWifiOnlyCell;
- (UITableViewCell *)preparedIncludeVideoCell;
- (UITableViewCell *)preparedDisableSleepCell;
@end


@implementation U1SettingsViewController

@synthesize navBar, tableView;
@synthesize autoUploadToggleCell, autoUploadToggle;
@synthesize wifiOnlyToggleCell, wifiOnlyToggleLabel, wifiOnlyToggle;
@synthesize includeVideoToggleCell, includeVideoToggleLabel, includeVideoToggle;
@synthesize disableSleepToggleCell, disableSleepToggleLabel, disableSleepToggle;
@synthesize accountLoadingCell, filesClient, accountInfo;

- (id)initWithNibName:(NSString *)nibNameOrNil bundle:(NSBundle *)nibBundleOrNil
{
    self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil];
    if (self) {
		self.title = NSLocalizedString(@"Settings", @"");
		self.filesClient = [U1FilesClient sharedFilesClient];
    }
    return self;
}

- (void)dealloc;
{
	[navBar release];
	[tableView release];
	[autoUploadToggleCell release];
	[autoUploadToggle release];
	[wifiOnlyToggleCell release];
	[wifiOnlyToggleLabel release];
	[wifiOnlyToggle release];
	[includeVideoToggleCell release];
	[includeVideoToggleLabel release];
	[includeVideoToggle release];
	[disableSleepToggleCell release];
	[disableSleepToggleLabel release];
	[disableSleepToggle release];
	[filesClient release];
	[accountInfo release];
	[super dealloc];
}

- (UITableViewCell *)accountLoadingCell;
{
	if (accountLoadingCell == nil)
	{
		accountLoadingCell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:nil];
		[accountLoadingCell.textLabel setText:NSLocalizedString(@"Loading Account Details...", nil)];
	}
	return accountLoadingCell;
}

- (void)viewWillAppear:(BOOL)animated;
{
	[super viewWillAppear:animated];
	[self.filesClient accountInfoWithCompletionBlock:^(NSDictionary *fetchedAccountInfo, NSError *error) {
		self.accountInfo = fetchedAccountInfo;
		if (self.accountInfo)
			[self.tableView reloadSections:[NSIndexSet indexSetWithIndex:U1SettingsViewSectionAccount] withRowAnimation:UITableViewRowAnimationNone];
	}];
}

- (IBAction)signOut:(id)sender;
{
	// TODO: get actual copy for this from design team
	UIActionSheet *actionSheet = [[UIActionSheet alloc] initWithTitle:@"Signing out on this device will not affect other devices connected to Ubuntu One." delegate:self cancelButtonTitle:@"Cancel" destructiveButtonTitle:@"Sign Out" otherButtonTitles:nil];
	[actionSheet showFromTabBar:self.tabBarController.tabBar];
	[actionSheet release];
}

- (void)actionSheet:(UIActionSheet *)actionSheet didDismissWithButtonIndex:(NSInteger)buttonIndex;
{
	if ([actionSheet destructiveButtonIndex] == buttonIndex)
	{
		[[U1AccountManager sharedAccountManager] removeCredentials];
		NSError *error;
		if(![[U1DataRepository sharedDataRepository] reset:&error]){
			CLS_LOG(@"Error resetting cache: %@", error);
		}
		
	}
}

- (void)toggleAutoUpload:(id)sender;
{
    BOOL switchValue = ((UISwitch *)sender).on;
    if ([CLLocationManager locationServicesEnabled] == YES )
    {
        CLAuthorizationStatus auth = [CLLocationManager authorizationStatus];
        if (auth == kCLAuthorizationStatusAuthorized ||
            auth == kCLAuthorizationStatusNotDetermined)
        {
            [[NSUserDefaults standardUserDefaults] setBool:switchValue forKey:@"auto_upload"];
            if (switchValue)
            {
                self.includeVideoToggleLabel.textColor = [UIColor blackColor];
                self.includeVideoToggle.enabled = YES;
                self.wifiOnlyToggleLabel.textColor = [UIColor blackColor];
                self.wifiOnlyToggle.enabled = YES;
                self.disableSleepToggleLabel.textColor = [UIColor blackColor];
                self.disableSleepToggle.enabled = YES;
            }
            else
            {
                self.includeVideoToggleLabel.textColor = [UIColor grayColor];
                self.includeVideoToggle.enabled = NO;
                self.wifiOnlyToggleLabel.textColor = [UIColor grayColor];
                self.wifiOnlyToggle.enabled = NO;
                self.disableSleepToggleLabel.textColor = [UIColor grayColor];
                self.disableSleepToggle.enabled = NO;
            }
        }
        else
        {
            UIAlertView *alert=[[[UIAlertView alloc] initWithTitle:@"Location Services Off" message:@"U1 Files app must be authorized to use location services to upload from Photos. In Settings App turn on Locatation Services -> U1Files" delegate:nil cancelButtonTitle:@"Ok" otherButtonTitles:nil] autorelease];
            [alert show];
            [(UISwitch *)sender setOn:NO animated:NO];
        }
    }
    else
    {
        // present error dialog with directions how to turn it on
        UIAlertView *alert=[[[UIAlertView alloc] initWithTitle:@"Location Services Off" message:@"Location services must be on to upload from Photos. In Settings App turn on Locatation Services" delegate:nil cancelButtonTitle:@"Ok" otherButtonTitles:nil] autorelease];
        [alert show];
        [(UISwitch *)sender setOn:NO animated:NO];
    }
}

- (void)toggleWifiOnly:(id)sender;
{
	BOOL switchValue = ((UISwitch *)sender).on;
	[[NSUserDefaults standardUserDefaults] setBool:switchValue forKey:@"wifi_only"];
}

- (void)toggleIncludeVideo:(id)sender;
{
	BOOL switchValue = ((UISwitch *)sender).on;
	[[NSUserDefaults standardUserDefaults] setBool:switchValue forKey:@"include_video"];
}

- (void)toggleDisableSleep:(id)sender;
{
	BOOL switchValue = ((UISwitch *)sender).on;
	[[NSUserDefaults standardUserDefaults] setBool:switchValue forKey:@"disable_sleep"];
}

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView;
{
	return U1SettingsViewSectionCount;
}

- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section;
{
	switch (section)
	{
		case U1SettingsViewSectionAccount:
			return NSLocalizedString(@"Account Details", nil);
		case U1SettingsViewSectionCameraRoll:
			return NSLocalizedString(@"Auto-Upload Camera Roll", nil);
	}
	return nil;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section;
{
	switch (section)
	{
		case U1SettingsViewSectionAccount:
			return 2;
		case U1SettingsViewSectionCameraRoll:
			return 3; // TODO const
	}
	return 0;
}

- (CGFloat)tableView:(UITableView *)tableView heightForRowAtIndexPath:(NSIndexPath *)indexPath;
{
	if (indexPath.section == U1SettingsViewSectionCameraRoll &&
		indexPath.row == 2)
	{
		return 70.0f;
	}
	return 44.0f;
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath;
{
	UITableViewCell *cell = nil;
	switch (indexPath.section)
	{
		case U1SettingsViewSectionAccount:
		{
			cell = [self.tableView dequeueReusableCellWithIdentifier:@"AccountCell"];
			if (!cell)
			{
				cell = [[[UITableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:@"AccountCell"] autorelease];
				[cell setSelectionStyle:UITableViewCellSelectionStyleNone];
			}
			switch (indexPath.row)
			{
				case 0:
					[cell.textLabel setText:NSLocalizedString(@"Name", nil)];
					[cell.detailTextLabel setText:@"Loading..."];
					if (self.accountInfo)
						[cell.detailTextLabel setText:[self.accountInfo objectForKey:@"visible_name"]];
					break;
				case 1:
				{
					[cell.textLabel setText:NSLocalizedString(@"Storage", nil)];
					U1ByteSizeValueTransformer *byteSizeTransformer = [[U1ByteSizeValueTransformer alloc] init];
					NSString *storageFormat = NSLocalizedString(@"%@ of %@ used", nil);
					NSString *usedString = [byteSizeTransformer transformedValue:[self.accountInfo objectForKey:@"used_bytes"]];
					NSString *totalString = [byteSizeTransformer transformedValue:[self.accountInfo objectForKey:@"max_bytes"]];
					[cell.detailTextLabel setText:@"Loading..."];
					if (self.accountInfo)
						[cell.detailTextLabel setText:[NSString stringWithFormat:storageFormat, usedString, totalString]];
					[byteSizeTransformer release];
					break;
				}
			}
			break;
//			return cell;
		}
		case U1SettingsViewSectionCameraRoll:
		{
			switch (indexPath.row)
			{
				case 0: cell = [self preparedAutoUploadCell]; break;
				case 1: cell = [self preparedWifiOnlyCell]; break;
//				case 2: cell = [self preparedIncludeVideoCell]; break;
				case 2: cell = [self preparedDisableSleepCell]; break;
				default: break;
			}
			[cell setSelectionStyle:UITableViewCellSelectionStyleNone];
		}
	}
	return cell;
}

- (NSIndexPath *)tableView:(UITableView *)tableView willSelectRowAtIndexPath:(NSIndexPath *)indexPath;
{
	return nil;
}

- (CGFloat)tableView:(UITableView *)tableView heightForFooterInSection:(NSInteger)section;
{
	if (section == U1SettingsViewSectionAccount)
		return 60.f;
	return 0.f;
}

- (UIView *)tableView:(UITableView *)tableView viewForFooterInSection:(NSInteger)section;
{
	if (section == U1SettingsViewSectionAccount)
	{
		CGRect footerBounds = CGRectMake(0, 0, 320, 60);
		UIView *footerView = [[[UIView alloc] initWithFrame:footerBounds] autorelease];
		
		UIButton *signOutButton = [[[UIButton alloc] initWithFrame:CGRectMake(10, 8, 300, 38)] autorelease];
		[footerView addSubview:signOutButton];
		UIImage *buttonImage = [[UIImage imageNamed:@"ios_files_buttonBG"] stretchableImageWithLeftCapWidth:0 topCapHeight:0];
		[signOutButton setBackgroundImage:buttonImage forState:UIControlStateNormal];
		[signOutButton setTitle:@"Sign Out of Ubuntu One" forState:UIControlStateNormal];
		[signOutButton.titleLabel setFont:[UIFont boldSystemFontOfSize:14.f]];
		[signOutButton.layer setCornerRadius:8.f];
		[signOutButton setClipsToBounds:YES];
		[signOutButton addTarget:self action:@selector(signOut:) forControlEvents:UIControlEventTouchUpInside];
		return footerView;
	}
	return nil;
}

- (UITableViewCell *)preparedAutoUploadCell;
{
	self.autoUploadToggle.on = [[NSUserDefaults standardUserDefaults] boolForKey:@"auto_upload"];
	return self.autoUploadToggleCell;
}

- (UITableViewCell *)preparedWifiOnlyCell;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];

	BOOL enabled = [defaults boolForKey:@"auto_upload"];
	
	if (enabled)
	{
		self.wifiOnlyToggleLabel.textColor = [UIColor blackColor];
		self.wifiOnlyToggle.enabled = YES;
	}
	else
	{
		self.wifiOnlyToggleLabel.textColor = [UIColor grayColor];
		self.wifiOnlyToggle.enabled = NO;
	}

	self.wifiOnlyToggle.on = [defaults boolForKey:@"wifi_only"];
	
	return self.wifiOnlyToggleCell;
}

- (UITableViewCell *)preparedIncludeVideoCell;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	
	BOOL enabled = [defaults boolForKey:@"auto_upload"];
	
	if (enabled)
	{
		self.includeVideoToggleLabel.textColor = [UIColor blackColor];
		self.includeVideoToggle.enabled = YES;
	}
	else
	{
		self.includeVideoToggleLabel.textColor = [UIColor grayColor];
		self.includeVideoToggle.enabled = NO;
	}
	
	self.includeVideoToggle.on = [defaults boolForKey:@"include_video"];
	
	return self.includeVideoToggleCell;
}

- (UITableViewCell *)preparedDisableSleepCell;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	
	BOOL enabled = [defaults boolForKey:@"auto_upload"];
	
	if (enabled)
	{
		self.disableSleepToggleLabel.textColor = [UIColor blackColor];
		self.disableSleepToggle.enabled = YES;
	}
	else
	{
		self.disableSleepToggleLabel.textColor = [UIColor grayColor];
		self.disableSleepToggle.enabled = NO;
	}
	
	self.disableSleepToggle.on = [defaults boolForKey:@"disable_sleep"];
	
	return self.disableSleepToggleCell;
}

@end
