//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1SingleSignOnProvider.h"

#import "U1Logging.h"

#import "U1HTTPRequestOperation.h"
#import "U1SSOAccountResponse.h"
#import "U1SSOAuthenticateResponse.h"
#import "U1SSOCaptchaResponse.h"


static NSString * const UOSingleSignOnHost = @"login.ubuntu.com";
static NSString * const UOSingleSignOnRoot_1_0 = @"/api/1.0";

// TODO: these can be returned from the server by querying the API Root
static NSString * const UOSingleSignOnPathAccounts = @"/accounts";
static NSString * const UOSingleSignOnPathAuthentications = @"/authentications";
static NSString * const UOSingleSignOnPathCaptchas = @"/captchas";
static NSString * const UOSingleSignOnPathRegistration = @"/registration";


static NSString *U1SSOEncodedStringFromParametersDictionary(NSDictionary *arguments);


@interface U1SingleSignOnProvider ()
@property (retain) NSOperationQueue *networkQueue;
- (void)sendGetToPath:(NSString*)path withParams:(NSString*)params credential:(NSURLCredential*)credential completionHandler:(void(^)(NSDictionary *result, NSError *error))handler;
- (void)sendPostToPath:(NSString*)path withParams:(NSString*)params credential:(NSURLCredential*)credential completionHandler:(void(^)(NSDictionary *result, NSError *error))handler;
@end


@implementation U1SingleSignOnProvider

@synthesize delegate;
@synthesize networkQueue;

- (id)initWithNetworkQueue:(NSOperationQueue*)aNetworkQueue;
{
	if (!(self = [super init]))
		return nil;
	
	// TODO: really the SSO provider should have a network service which allows it to
	// GET/POST as necessary and not have to manage operations directly.
	self.networkQueue = aNetworkQueue ?: [[[NSOperationQueue alloc] init] autorelease];
	
	return self;
}

- (void)dealloc;
{
	self.delegate = nil;
	self.networkQueue = nil;
	[super dealloc];
}

- (id)requestNewCaptchaWithCompletionHandlerOnMainQueue:(void(^)(U1SSOCaptchaResponse *response, NSError *error))handler;
{
	NSString *params = @"ws.op=new";
	[self sendPostToPath:UOSingleSignOnPathCaptchas withParams:params credential:nil completionHandler:^(NSDictionary *result, NSError *error) {
		U1SSOCaptchaResponse *captchaResponse = nil;
		if (result != nil && error == nil)
		{
			captchaResponse = [[[[U1SSOCaptchaResponse alloc] init] autorelease] parseFromJSONDictionary:result];
		}
		dispatch_async(dispatch_get_main_queue(), ^{
			handler(captchaResponse, error);
		});
	}];
	return nil;
}

- (id)registerAccountForEmail:(NSString*)email
				 withPassword:(NSString*)password
				  displayName:(NSString*)displayName
					captchaId:(NSString*)captchaId
			  captchaSolution:(NSString*)captchaSolution
			completionHandlerOnMainQueue:(void(^)(U1SSOResponse *response, NSError *error))handler;
{
	NSMutableDictionary *paramsDict = [NSMutableDictionary dictionary];
	[paramsDict setObject:@"register" forKey:@"ws.op"];
	[paramsDict setObject:email forKey:@"email"];
	[paramsDict setObject:password forKey:@"password"];
	[paramsDict setObject:captchaId forKey:@"captcha_id"];
	[paramsDict setObject:captchaSolution forKey:@"captcha_solution"];
	[paramsDict setObject:@"mobile" forKey:@"platform"];
	[paramsDict setObject:@"https://one.ubuntu.com/oauth/complete-registration/?platform=mobile&return_to=x-ubuntuone-sso-files://signup/complete" forKey:@"validate_redirect_to"];
	if (displayName)
		[paramsDict setObject:displayName forKey:@"displayname"];	
	NSString *params = U1SSOEncodedStringFromParametersDictionary(paramsDict);
	
	[self sendPostToPath:UOSingleSignOnPathRegistration withParams:params credential:nil completionHandler:^(NSDictionary *result, NSError *error) {
		
		U1SSOResponse *response = nil;
		if (!error)
		{
			response = [[[[U1SSOResponse alloc] init] autorelease] parseFromJSONDictionary:result];
		}
        dispatch_async(dispatch_get_main_queue(), ^{
            handler(response, error);
        });
		
	}];
	return nil;
}

- (id)authenticateAccountForEmail:(NSString*)email
					 withPassword:(NSString*)password
						 appToken:(NSString*)appToken
				completionHandlerOnMainQueue:(void(^)(U1SSOAuthenticateResponse *response, NSError *error))handler;
{
	NSDictionary *paramsDict = [NSDictionary dictionaryWithObjectsAndKeys:
								@"authenticate", @"ws.op",
								appToken, @"token_name",
								nil];
	NSString *params = U1SSOEncodedStringFromParametersDictionary(paramsDict);
	NSURLCredential *credential = [NSURLCredential credentialWithUser:email password:password persistence:NSURLCredentialPersistenceNone];
	[self sendGetToPath:UOSingleSignOnPathAuthentications withParams:params credential:credential completionHandler:^(NSDictionary *result, NSError *error){
		U1SSOAuthenticateResponse *authResponse = nil;
		if (result)
		{
			authResponse = [[[[U1SSOAuthenticateResponse alloc] init] autorelease] parseFromJSONDictionary:result];
		}
        dispatch_async(dispatch_get_main_queue(), ^{
            handler(authResponse, error);
        });
	}];
	return nil;
}

- (id)requestPasswordResetTokenForEmail:(NSString*)email withCompletionHandlerOnMainQueue:(void(^)(U1SSOResponse *response, NSError *error))handler;
{
	NSDictionary *paramsDict = [NSDictionary dictionaryWithObjectsAndKeys:
								@"request_password_reset_token", @"ws.op",
								email, @"email",
								nil];
	NSString *params = U1SSOEncodedStringFromParametersDictionary(paramsDict);
	[self sendPostToPath:UOSingleSignOnPathRegistration withParams:params credential:nil completionHandler:^(NSDictionary *result, NSError *error) {
		U1SSOResponse *response = nil;
		if (result)
		{
			response = [[[[U1SSOResponse alloc] init] autorelease] parseFromJSONDictionary:result];
		}
        dispatch_async(dispatch_get_main_queue(), ^{
            handler(response, error);
        });
    }];
	return nil;
}

- (id)setNewPassword:(NSString*)password forEmail:(NSString*)email withToken:(NSString*)token completionHandlerOnMainQueue:(void(^)(U1SSOResponse *response, NSError *error))handler;
{
	NSDictionary *paramsDict = [NSDictionary dictionaryWithObjectsAndKeys:
								@"set_new_password", @"ws.op",
								email, @"email",
								password, @"newPassword",
								token, @"token",
								nil];
	NSString *params = U1SSOEncodedStringFromParametersDictionary(paramsDict);
	[self sendPostToPath:UOSingleSignOnPathRegistration withParams:params credential:nil completionHandler:^(NSDictionary *result, NSError *error) {
		U1SSOResponse *response = nil;
		if (result)
		{
			response = [[[[U1SSOResponse alloc] init] autorelease] parseFromJSONDictionary:result];
		}
        dispatch_async(dispatch_get_main_queue(), ^{
            handler(response, error);
        });
	}];
	return nil;
}

- (id)requestAccountDetailsWithCompletionHandler:(void(^)(NSError *error))handler;
{
	
	return nil;
}


#pragma mark Private Methods

- (void)sendGetToPath:(NSString*)path withParams:(NSString*)params credential:(NSURLCredential*)credential completionHandler:(void(^)(NSDictionary *result, NSError *error))handler;
{
	// TODO: extract scheme/host into initialization parameters
	NSString *urlString = [@"https://" stringByAppendingFormat:@"%@%@%@?%@", UOSingleSignOnHost, UOSingleSignOnRoot_1_0, path, params];
	NSURL *url = [NSURL URLWithString:urlString];
	NSURLRequest *request = [NSURLRequest requestWithURL:url];
	U1HTTPRequestOperationInMemoryDataCollector *collector = [[[U1HTTPRequestOperationInMemoryDataCollector alloc] init] autorelease];
	U1HTTPRequestOperation *httpOp = [[[U1HTTPRequestOperation alloc] initWithRequest:request dataCollector:collector] autorelease];
	httpOp.credential = credential;
	[httpOp setCompletionBlock:^(void) {
		NSError *error = httpOp.error;
		NSDictionary *body = nil;
		if (!error)
		{
			NSHTTPURLResponse *response = httpOp.response;
			if ([response statusCode] == 200) {
                body = [NSJSONSerialization JSONObjectWithData:collector.data options:0 error:&error];
                if(body == nil){
                    CLS_LOG(@"failed to parse response: %@", error);
                }
            }else{
                // TODO: handle non-200 response errors
                CLS_LOG(@"error: GET returned status %ld", (long)[response statusCode]);
            }
		}
		handler(body, error);
	}];
	[self.networkQueue addOperation:httpOp];
}

- (void)sendPostToPath:(NSString*)path withParams:(NSString*)params credential:(NSURLCredential*)credential completionHandler:(void(^)(NSDictionary *result, NSError *error))handler;
{
	// TODO: extract scheme/host into initialization parameters
	NSString *urlString = [@"https://" stringByAppendingFormat:@"%@%@%@", UOSingleSignOnHost, UOSingleSignOnRoot_1_0, path];
	NSURL *url = [NSURL URLWithString:urlString];
	NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:url];
	[request setHTTPMethod:@"POST"];
	[request setHTTPBody:[params dataUsingEncoding:NSUTF8StringEncoding]];
	[request setValue:@"application/x-www-form-urlencoded" forHTTPHeaderField:@"Content-Type"];
	
	U1HTTPRequestOperationInMemoryDataCollector *collector = [[[U1HTTPRequestOperationInMemoryDataCollector alloc] init] autorelease];
	U1HTTPRequestOperation *httpOp = [[[U1HTTPRequestOperation alloc] initWithRequest:request dataCollector:collector] autorelease];
	httpOp.credential = credential;
	[httpOp setCompletionBlock:^(void) {
		NSError *error = httpOp.error;
		NSDictionary *body = nil;
		if (!error)
		{
			NSHTTPURLResponse *response = httpOp.response;
			if ([response statusCode] == 200) {
				body = [NSJSONSerialization JSONObjectWithData:collector.data options:0 error:&error];
				if(body == nil){
					CLS_LOG(@"failed to parse response: %@", error);
				}
			}else{
				// TODO: handle non-200 response errors
				CLS_LOG(@"error: GET returned status %ld", (long)[response statusCode]);
			}
		}
		handler(body, error);
	}];
	[self.networkQueue addOperation:httpOp];
}

@end


static NSString *U1SSOEncodedStringFromParametersDictionary(NSDictionary *arguments)
{
	NSMutableArray *encodedArguments = [NSMutableArray array];
	[arguments enumerateKeysAndObjectsUsingBlock:^(NSString *key, NSString *value, BOOL *stop) {
        if (![key isEqualToString:@"ws.op"]) {
            value = [NSString stringWithFormat:@"\"%@\"", value];
        }
		value = [(id)CFURLCreateStringByAddingPercentEscapes(kCFAllocatorDefault, (CFStringRef)value, NULL, (CFStringRef)@":/?#[]@!$&’()*+,;=", kCFStringEncodingUTF8) autorelease];
		NSString *argumentPair = [key stringByAppendingFormat:@"=%@", value];
		[encodedArguments addObject:argumentPair];
	}];

	return [encodedArguments componentsJoinedByString:@"&"];
}
