//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1UploadOperation.h"

#import "U1Logging.h"

#import "U1Asset.h"
#import "U1AssetRepresenationDataProvider.h"
#import "U1DataRepository.h"
#import "U1FilesService.h"
#import "U1FileNode.h"
#import "U1FolderNode.h"
#import "U1LocalFileInfo.h"
#import "U1MagicUploadDataProvider.h"


@interface U1UploadOperation ()
@property (getter=isExecuting) BOOL executing;
@property (getter=isFinished) BOOL finished;
- (void)finishExecuting;
- (void)beginUpload;
@end


@implementation U1UploadOperation

@synthesize executing, finished, assetID, mimetype, error, progressBlock, fileInfo;

- (id)init;
{
	self = [super init];
	if (!self)
		return nil;
	fileInfo = [[U1LocalFileInfo alloc] init];
	return self;
}

- (BOOL)isConcurrent;
{
	return YES;
}

- (void)finishExecuting;
{
	[self willChangeValueForKey:@"isExecuting"];
	[self willChangeValueForKey:@"isFinished"];
	self.finished = YES;
	self.executing = NO;
	[self didChangeValueForKey:@"isFinished"];
	[self didChangeValueForKey:@"isExecuting"];
//	[UONetworkStatusCoordinator removeNetworkActivity];
}

- (void)start;
{
	[self willChangeValueForKey:@"isExecuting"];
	self.executing = YES;
	[self didChangeValueForKey:@"isExecuting"];

	[self beginUpload];
}

- (void)beginUpload;
{
	if ([self isCancelled])
	{
		[self finishExecuting];
		return;
	}
	
	dispatch_async(dispatch_get_main_queue(), ^(void) {

		[self.fileInfo setUploadPercentage:0.0];
		[self.fileInfo setUploading:YES];
		[self.fileInfo setUploaded:NO];

		U1FilesService *filesService = [U1FilesService sharedFilesService];
		fileInfo.uploading = YES;

		U1MagicUploadDataProvider *dataProvider = [[[U1MagicUploadDataProvider alloc] init] autorelease];
		dataProvider.assetID = self.assetID;
		
		[filesService uploadContentData:dataProvider
						  progressBlock:^(long long uploaded, long long totalLength) {}
			 completionBlockOnMainQueue:^(NSDictionary *updatedNodeInfo, NSError *magicUploadError) {
			
			if (magicUploadError != nil)
			{
				if ([magicUploadError.domain isEqualToString:@"nonexistentAsset"])
				{
					self.error = magicUploadError;
					// Asset is gone, delete U1Asset and potentially the Node, and stop running

					U1DataRepository *dataRepository = [U1DataRepository sharedDataRepository];
					[dataRepository dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
						U1Asset *assetToBeDeleted = (U1Asset*)[context objectWithID:self.assetID];
						U1FileNode *fileNodeToBeDeleted = assetToBeDeleted.fileNode;
						[context deleteObject:assetToBeDeleted];
						if (fileNodeToBeDeleted.generation == nil)
						{
							[context deleteObject:fileNodeToBeDeleted];
						}
						NSError *saveError;
						if(![context save:&saveError]){
							CLS_LOG(@"error saving deletion of placeholder for asset of node: err=%@", saveError);
						}
						
					}];
					[self finishExecuting];
				}
				else
				{
					NSLog(@"magic uploadError != nil and domain isn't nonexistentAsset, it's %@, code %d", magicUploadError.domain, magicUploadError.code);
					// Magic upload failed, try a standard upload
					U1AssetRepresenationDataProvider *newDataProvider = [[[U1AssetRepresenationDataProvider alloc] init] autorelease];
					newDataProvider.assetID = self.assetID;
					
					[filesService uploadContentData:newDataProvider progressBlock:^(long long uploaded, long long totalLength) {
						dispatch_async(dispatch_get_main_queue(), ^{
							[self.fileInfo setUploadPercentage:((double)uploaded / (double)totalLength)];
						});
						
						if (self.progressBlock)
							self.progressBlock(uploaded, totalLength);
						
					} completionBlockOnMainQueue:^(NSDictionary *updatedNodeInfo, NSError *regularUploadError) {
						if (regularUploadError != nil)
						{
							self.error = regularUploadError;
						}

						[self.fileInfo setUploadPercentage:1.0];
						[self.fileInfo setUploading:NO];
						[self.fileInfo setUploaded:YES];
						
						[[U1DataRepository sharedDataRepository] dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
							U1Asset *asset = (U1Asset *)[context objectWithID:self.assetID];
							[asset.fileNode updatePropertiesFromJSONDictionary:updatedNodeInfo];
							NSError *saveError;
							if(![context save:&saveError]){
								CLS_LOG(@"error saving updated properties for successfully uploaded node: err=%@", saveError);
							}
						}];
						[self finishExecuting];
					}];
				}
			}
			else
			{
				[self.fileInfo setUploadPercentage:1.0];
				[self.fileInfo setUploading:NO];
				[self.fileInfo setUploaded:YES];
				
				[[U1DataRepository sharedDataRepository] dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
					U1Asset *asset = (U1Asset *)[context objectWithID:self.assetID];
					[asset.fileNode updatePropertiesFromJSONDictionary:updatedNodeInfo];
					NSError *saveError;
					if(![context save:&saveError]){
						CLS_LOG(@"error saving updated properties for successfully magic-uploaded node: err=%@", saveError);
					}
					NSLog(@"successfully magic-uploaded and saved updated info for asset:\n%@", asset);
					
				}];
				[self finishExecuting];
			}
		}];
	});
}

- (void)dealloc
{
	[mimetype release];
	[error release];
	[progressBlock release];
	[fileInfo release];
    [super dealloc];
}

@end
