//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1UploadsPoolViewController.h"

#import "U1Logging.h"

#import "U1FileNode.h"
#import "U1FileDetailsViewController.h"
#import "U1FilesClient.h"
#import "U1FolderItemCell.h"
#import "U1FolderNode.h"
#import "U1Asset.h"
#import "U1Node.h"
#import "UIActionSheet+U1Additions.h"
#import "U1DataRepository.h"
#import "U1LocalFileInfo.h"


@interface U1UploadsPoolViewController () <UINavigationControllerDelegate, UIImagePickerControllerDelegate, NSFetchedResultsControllerDelegate>
@property (retain) NSFetchedResultsController *resultsController;
- (void)updateFeedback:(id)sender;
- (void)loadCurrentUploads;
- (void)reloadTableData:(id)sender;
- (UIView *)preparedAutoUploadToggleView;
- (void)takePicture:(id)sender;
- (void)configureCell:(U1FolderItemCell*)cell forAsset:(U1Asset*)asset withInfo:(U1LocalFileInfo*)fileInfo;
@end


@implementation U1UploadsPoolViewController

@synthesize loadingCell;
@synthesize autoUploadToggleView, autoUploadToggle;
@synthesize filesClient;
@synthesize resultsController;

- (id)init;
{
	self = [super initWithNibName:@"U1UploadsPoolViewController" bundle:nil];
	if (!self)
		return nil;
	self.title = NSLocalizedString(@"Uploads", @"");
	filesClient = [[U1FilesClient sharedFilesClient] retain];
	[[NSNotificationCenter defaultCenter] addObserver:self
											 selector:@selector(reloadTableData:) name:NSUserDefaultsDidChangeNotification object:nil];
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(updateFeedback:) name:@"uploadQueued" object:nil];
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(updateFeedback:) name:@"imageUploaded" object:nil];

	return self;
}

- (void)dealloc;
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:NSUserDefaultsDidChangeNotification object:nil];
	[[NSNotificationCenter defaultCenter] removeObserver:self name:@"imageUploaded" object:nil];
	[filesClient release];
	[loadingCell release];
	[resultsController release];
	[autoUploadToggleView release];
	[autoUploadToggle release];
	[super dealloc];
}

- (void)viewDidLoad;
{
	[super viewDidLoad];
	[self updateFeedback:nil];
	[self loadCurrentUploads];
}

- (void)viewDidUnload;
{
	[loadingCell release], loadingCell = nil;
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	[super viewDidUnload];
}

- (void)viewDidAppear:(BOOL)animated;
{
	[super viewDidAppear:animated];
	[self.tableView deselectRowAtIndexPath:[self.tableView indexPathForSelectedRow] animated:YES];
	
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(nodeStatusChanged:) name:U1FilesClientNodeStatusChangedNotification object:nil];
}

- (void)viewWillDisappear:(BOOL)animated;
{
	[super viewWillDisappear:animated];
	[[NSNotificationCenter defaultCenter] removeObserver:self name:U1FilesClientNodeStatusChangedNotification object:nil];
}

- (void)nodeStatusChanged:(NSNotification*)notification;
{
	U1FileNode *aNode = (id)[[notification userInfo] objectForKey:@"node"];
	U1LocalFileInfo *fileInfo = [[notification userInfo] objectForKey:@"fileInfo"];
	U1Asset *asset = aNode.asset;
	
	NSIndexPath *indexPath = [self.resultsController indexPathForObject:asset];
	U1FolderItemCell *cell = (id)[self.tableView cellForRowAtIndexPath:indexPath];
	if (cell)
	{
		[self configureCell:cell forAsset:asset withInfo:fileInfo];
	}
}

- (void)updateFeedback:(id)sender;
{
	// Update the label and the progress bar
	// progressLabel = (numberOfAssets - assetsPendingUpload) / numberOfAssets
	dispatch_async(dispatch_get_main_queue(), ^{
		NSNumberFormatter *countFormatter = [[NSNumberFormatter alloc] init];
		[countFormatter setNumberStyle:NSNumberFormatterDecimalStyle];
		NSNumber *assetsToUpload = [[NSNumber alloc] initWithInt:[self.filesClient queueCount]];
		NSString *badgeValue = nil;
		if ([assetsToUpload intValue] > 0)
		{
			badgeValue = [countFormatter stringFromNumber:assetsToUpload];
		}
		self.navigationController.tabBarItem.badgeValue = badgeValue;
		[countFormatter release];
		[assetsToUpload release];
		
		// Maybe disable sleep
		BOOL disableSleep = ([[NSUserDefaults standardUserDefaults] boolForKey:@"disable_sleep"] &&
							 [self.filesClient isOperating]);
		if ([[UIDevice currentDevice] batteryState] == UIDeviceBatteryStateUnplugged)
		{
			// Absolutely do not disable sleep, no matter what.
			disableSleep = NO;
		}
		[[UIApplication sharedApplication] setIdleTimerDisabled:disableSleep];

		[self.tableView reloadData];
	});
}

- (void)takePicture:(id)sender;
{
	UIImagePickerController *imagePicker = [[UIImagePickerController alloc] init];
	imagePicker.sourceType = UIImagePickerControllerSourceTypeCamera;
	imagePicker.delegate = self;
	[self presentModalViewController:imagePicker animated:YES];
}

- (void)imagePickerController:(UIImagePickerController *)picker didFinishPickingMediaWithInfo:(NSDictionary *)info;
{
	[picker dismissModalViewControllerAnimated:YES];
	
	UIImage *image = [info objectForKey:UIImagePickerControllerOriginalImage];
	
	UIImageWriteToSavedPhotosAlbum(image, self, nil, nil);
}

#pragma mark UITableViewDataSource Methods

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section;
{
	if (!self.resultsController)
		return 1;
	
	return [[[self.resultsController sections] objectAtIndex:section] numberOfObjects];
}

- (CGFloat)tableView:(UITableView *)tableView heightForHeaderInSection:(NSInteger)section;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	
	if (section == 0 &&
		(![defaults boolForKey:@"auto_upload"]))
	{
		return 54.0f;
	}
	
	return 0.0f;
}

- (UIView *)tableView:(UITableView *)tableView viewForHeaderInSection:(NSInteger)section;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	
	if (section == 0)
	{
		if (![defaults boolForKey:@"auto_upload"])
		{
			return [self preparedAutoUploadToggleView];
		}
	}
	
	return nil;
}

- (void)configureCell:(U1FolderItemCell*)cell forAsset:(U1Asset*)asset withInfo:(U1LocalFileInfo*)fileInfo;
{
	[cell setAccessoryType:UITableViewCellAccessoryNone];
	cell.nameLabel.text = asset.filename;
	cell.nameLabel.textColor = [UIColor blackColor];
	NSString *detailString = [NSString stringWithFormat:@"%@ %@", NSLocalizedString(@"Pending upload to", @""), [asset.fileNode.parent name]];
	
	//  TODO: get this thumbnail loading happening inside an operation queue, tied to visible rows
	//	[self.autoUploadsManager thumbnailForNode:childNode completionBlock:^(CGImageRef thumbnail) {
	//		cell.iconImageView.image = [UIImage imageWithCGImage:thumbnail];
	//	}];
	
	// query the local assets manager to see if this childNode's corresponding asset is actually in the upload operation queue
	BOOL isUploading = [self.filesClient isUploading:asset.fileNode];
	if (isUploading)
	{
		[cell.uploadIndicator setHidden:NO];
		[cell.uploadIndicator startAnimating];
		detailString = [NSString stringWithFormat:@"%@ %@", NSLocalizedString(@"Uploading to", @""), [asset.fileNode.parent name]];
	}
	else {
		[cell.uploadIndicator setHidden:YES];
		[cell.uploadIndicator stopAnimating];
	}
	
	[cell.uploadProgressView setProgress:fileInfo.uploadPercentage];
	[cell.uploadProgressView setHidden:!fileInfo.isUploading];
	[cell.detailLabel setHidden:fileInfo.isUploading];
	
	UIDocumentInteractionController *docController = [UIDocumentInteractionController interactionControllerWithURL:[NSURL fileURLWithPath:asset.filename]];
	[cell.imageView setImage:[docController.icons lastObject]];
	[cell.detailLabel setText:detailString];
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath;
{
	if (!self.resultsController)
	{
		return self.loadingCell;
	}
	
	U1FolderItemCell *cell = [U1FolderItemCell cellForTableView:tableView];
	
	U1Asset *asset = [self.resultsController objectAtIndexPath:indexPath];

	[self configureCell:cell forAsset:asset withInfo:nil];
	
	return cell;
}


#pragma mark UITableViewDelegate Methods

- (void)reloadTableData:(id)sender;
{
	[self.tableView reloadData];
}

#pragma mark Private Methods

- (void)loadCurrentUploads;
{
	[[U1DataRepository sharedDataRepository] dispatchBlockWithMainContext:^(NSManagedObjectContext *context) {
		
		NSFetchRequest *request = [[NSFetchRequest alloc] init];
		[request setEntity:[U1Asset entityInManagedObjectContext:context]];
		NSPredicate *p = [NSPredicate predicateWithFormat:@"generation = %@", nil];
		[request setPredicate:p];
		NSArray *sortDescriptors = [NSArray arrayWithObjects:
									[NSSortDescriptor sortDescriptorWithKey:@"filename" ascending:YES selector:@selector(localizedStandardCompare:)],
									nil];
		[request setSortDescriptors:sortDescriptors];
		self.resultsController = [[[NSFetchedResultsController alloc] initWithFetchRequest:request managedObjectContext:context sectionNameKeyPath:nil cacheName:nil] autorelease];
		[request release];
	}];
	
	dispatch_async(dispatch_get_main_queue(), ^(void) {
		NSError *error = nil;
		if(![self.resultsController performFetch:&error]){
			CLS_LOG(@"error fetching current uploads: %@", error);
		}
		[self.resultsController setDelegate:self];
		[self.tableView reloadData];
	});
}

#pragma mark NSFetchedResultsControllerDelegate Methods


- (void)controllerWillChangeContent:(NSFetchedResultsController *)controller;
{
	[self.tableView beginUpdates];
}


- (void)controller:(NSFetchedResultsController *)controller didChangeSection:(id <NSFetchedResultsSectionInfo>)sectionInfo
		   atIndex:(NSUInteger)sectionIndex forChangeType:(NSFetchedResultsChangeType)type;
{
	switch (type)
	{
		case NSFetchedResultsChangeInsert:
			[self.tableView insertSections:[NSIndexSet indexSetWithIndex:sectionIndex]
						  withRowAnimation:UITableViewRowAnimationFade];
			break;
			
		case NSFetchedResultsChangeDelete:
			[self.tableView deleteSections:[NSIndexSet indexSetWithIndex:sectionIndex]
						  withRowAnimation:UITableViewRowAnimationFade];
			break;
	}
}

- (void)controller:(NSFetchedResultsController *)controller didChangeObject:(id)anObject
	   atIndexPath:(NSIndexPath *)indexPath forChangeType:(NSFetchedResultsChangeType)type
	  newIndexPath:(NSIndexPath *)newIndexPath;
{
	
	UITableView *tableView = self.tableView;
	
	switch (type)
	{
			
		case NSFetchedResultsChangeInsert:
			[tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:newIndexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			break;
			
		case NSFetchedResultsChangeDelete:
			[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			break;
			
		case NSFetchedResultsChangeUpdate:
			// TODO: 
			//            [self configureCell:[tableView cellForRowAtIndexPath:indexPath]
			//					atIndexPath:indexPath];
			[self.tableView reloadRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
								  withRowAnimation:UITableViewRowAnimationNone];
			break;
			
		case NSFetchedResultsChangeMove:
			[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			[tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:newIndexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			break;
	}
}

- (void)controllerDidChangeContent:(NSFetchedResultsController *)controller;
{
	[self.tableView endUpdates];
}

- (UIView *)preparedAutoUploadToggleView;
{
	self.autoUploadToggle.on = [[NSUserDefaults standardUserDefaults] boolForKey:@"auto_upload"];
	return self.autoUploadToggleView;
}

- (void)enableAutoUpload:(id)sender;
{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	[defaults setBool:((UISwitch *)sender).on forKey:@"auto_upload"];
	[self.tableView reloadData];
}

@end
