//
//  Copyright 2011-2013 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "U1VolumesViewController.h"

#import "U1Logging.h"

#import "U1FilesClient.h"
#import "U1FileDetailsViewController.h"
#import "U1Volume.h"
#import "U1FolderItemCell.h"
#import "U1FolderViewController.h"
#import "U1ByteSizeValueTransformer.h"


@interface U1VolumesViewController () <U1VolumesViewControllerDelegate, U1FolderViewControllerDelegate, NSFetchedResultsControllerDelegate>
@property (retain) NSFetchedResultsController *cloudFoldersFetchController;
@property (retain) U1ByteSizeValueTransformer *byteSizeTransformer;
- (void)loadVolumes;
@end


@implementation U1VolumesViewController

@synthesize loadingCell;
@synthesize delegate, filesClient;
@synthesize cloudFoldersFetchController, byteSizeTransformer;

- (id)init;
{
	self = [super initWithNibName:@"U1VolumesViewController" bundle:nil];
	if (!self)
		return nil;
	self.title = NSLocalizedString(@"Files", @"");
	filesClient = [[U1FilesClient sharedFilesClient] retain];
	self.byteSizeTransformer = [[U1ByteSizeValueTransformer new] autorelease];
	self.delegate = self; // ???
	return self;
}

- (void)dealloc;
{
	[filesClient release];
	[loadingCell release];
	[cloudFoldersFetchController release];
	[byteSizeTransformer release];
	[super dealloc];
}

- (void)viewDidLoad;
{
	[super viewDidLoad];
	[self loadVolumes];
}

- (void)refresh;
{
	[self loadVolumes];
}

- (void)viewDidUnload;
{
	[loadingCell release], loadingCell = nil;
	[super viewDidUnload];
}

- (void)viewDidAppear:(BOOL)animated;
{
	[super viewDidAppear:animated];
	[self.tableView deselectRowAtIndexPath:[self.tableView indexPathForSelectedRow] animated:YES];
}


#pragma mark UITableViewDataSource Methods

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView;
{
	return 1;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section;
{
	if (self.cloudFoldersFetchController == nil)
		return 1; // loading cell
	
	return [[[self.cloudFoldersFetchController sections] objectAtIndex:0] numberOfObjects];
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath;
{
	if (!self.cloudFoldersFetchController)
		return self.loadingCell;
	
	UITableViewCell *cell = [self.tableView dequeueReusableCellWithIdentifier:@"U1FolderCell"];
	if (!cell)
	{
		cell = [[[UITableViewCell alloc] initWithStyle:UITableViewCellStyleSubtitle reuseIdentifier:@"U1FolderCell"] autorelease];
		[cell setAccessoryType:UITableViewCellAccessoryDisclosureIndicator];
		[cell setSelectionStyle:UITableViewCellSelectionStyleGray];
	}
	U1Volume *volume = [self.cloudFoldersFetchController objectAtIndexPath:indexPath];
	[cell.imageView setImage:[UIImage imageNamed:@"ic_folder"]];
	if ([volume.kind isEqualToString:@"root"])
	{
		[cell.imageView setImage:[UIImage imageNamed:@"u1_folder"]];
	}
	NSString *volumePath = volume.path;
	if ([@"~/.ubuntuone/Purchased from Ubuntu One" isEqualToString:volumePath])
		volumePath = @"Purchased Music";
	if ([volumePath hasPrefix:@"~/"])
		volumePath = [volumePath substringFromIndex:2];
	cell.textLabel.text = volumePath;
	return cell;
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath;
{
	U1Volume *volume = [self.cloudFoldersFetchController objectAtIndexPath:indexPath];
	[self.delegate volumesController:self didSelectVolume:volume];
}



#pragma mark NSFetchedResultsControllerDelegate Methods


- (void)controllerWillChangeContent:(NSFetchedResultsController *)controller;
{
	[self.tableView beginUpdates];
}


- (void)controller:(NSFetchedResultsController *)controller didChangeSection:(id <NSFetchedResultsSectionInfo>)sectionInfo
		   atIndex:(NSUInteger)sectionIndex forChangeType:(NSFetchedResultsChangeType)type;
{
	switch (type)
	{
		case NSFetchedResultsChangeInsert:
			[self.tableView insertSections:[NSIndexSet indexSetWithIndex:sectionIndex]
						  withRowAnimation:UITableViewRowAnimationFade];
			break;
			
		case NSFetchedResultsChangeDelete:
			[self.tableView deleteSections:[NSIndexSet indexSetWithIndex:sectionIndex]
						  withRowAnimation:UITableViewRowAnimationFade];
			break;
	}
}

- (void)controller:(NSFetchedResultsController *)controller didChangeObject:(id)anObject
	   atIndexPath:(NSIndexPath *)indexPath forChangeType:(NSFetchedResultsChangeType)type
	  newIndexPath:(NSIndexPath *)newIndexPath;
{
	UITableView *tableView = self.tableView;
	
	switch (type)
	{
			
		case NSFetchedResultsChangeInsert:
			[tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:newIndexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			break;
			
		case NSFetchedResultsChangeDelete:
			[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			break;
			
		case NSFetchedResultsChangeUpdate:
			// TODO: 
			//            [self configureCell:[tableView cellForRowAtIndexPath:indexPath]
			//					atIndexPath:indexPath];
			[self.tableView reloadRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
								  withRowAnimation:UITableViewRowAnimationNone];
			break;
			
		case NSFetchedResultsChangeMove:
			[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			[tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:newIndexPath]
							 withRowAnimation:UITableViewRowAnimationFade];
			break;
	}
}

- (void)controllerDidChangeContent:(NSFetchedResultsController *)controller;
{
	[self.tableView endUpdates];
}


#pragma mark Private Methods

- (void)loadVolumes;
{
	[self.filesClient volumesWithResultsBlock:^(U1Volume *rootVolume, NSFetchedResultsController *cloudFoldersResultsController, U1DataSourceType resultsDataSourceType) {
		
		self.cloudFoldersFetchController = cloudFoldersResultsController;
		[self.cloudFoldersFetchController setDelegate:self];
		[self stopLoading];
		[self.tableView reloadData];
	} errorBlock:^(NSError *error) {
		CLS_LOG(@"Couldn't fetch remote volumes: %@", error);
	}];
}

#pragma mark U1VolumesControllerDelegate Methods

- (void)volumesController:(U1VolumesViewController *)volumesController didSelectVolume:(U1Volume *)volume;
{
	NSString *nodePath = volume.nodePath;
	U1FolderViewController *folderController = [[U1FolderViewController alloc] initWithNodeResourcePath:nodePath];
	folderController.delegate = self;
	[self.navigationController pushViewController:folderController animated:YES];
	[folderController release];
}

#pragma mark U1FolderViewControllerDelegate Methods

- (void)folderController:(U1FolderViewController*)folderController didSelectNode:(U1Node*)aNode;
{
	if ([aNode isFolder])
	{
		U1FolderViewController *nextFolderController = [[U1FolderViewController alloc] initWithNodeResourcePath:aNode.resourcePath];
		nextFolderController.delegate = self;
		[self.navigationController pushViewController:nextFolderController animated:YES];
		[nextFolderController release];
	}
	else if ([aNode isFile])
	{
		U1FileDetailsViewController *fileController = [[U1FileDetailsViewController alloc] initWithNode:(id)aNode];
		[self.navigationController pushViewController:fileController animated:YES];
		[fileController release];
	}
}

@end
