//
//  Copyright 2013 Canonical Ltd.
//
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


// Class under test
#import "U1AssetManager.h"
#import "U1Asset.h"

// Collaborators
#import "CoreDataTestCase.h"
#import <AssetsLibrary/AssetsLibrary.h>

// Test support
#import <SenTestingKit/SenTestingKit.h>

#define HC_SHORTHAND
#import <OCHamcrestIOS/OCHamcrestIOS.h>

#define MOCKITO_SHORTHAND
#import <OCMockitoIOS/OCMockitoIOS.h>

// Fake Asset Library that also implements the enumerateAssets method to fake ALAssetGroup.
@interface FakeAssetsLibrary : NSObject
@property (retain) NSMutableArray *mockAssets;
- (id)initWithMockAssetArray:(NSArray *)mockAssets;
- (void)enumerateGroupsWithTypes:(ALAssetsGroupType)types usingBlock:(ALAssetsLibraryGroupsEnumerationResultsBlock)enumerationBlock failureBlock:(ALAssetsLibraryAccessFailureBlock)failureBlock;
@end

@implementation FakeAssetsLibrary

- (id)initWithMockAssetArray:(NSArray *)newMockAssets{
	self = [super init];
	if(!self) return nil;
	self.mockAssets = [newMockAssets mutableCopy];
	return self;
}

- (void)addAsset:(ALAsset *)newAsset{
	[self.mockAssets addObject:newAsset];
}

- (void)enumerateGroupsWithTypes:(ALAssetsGroupType)types usingBlock:(ALAssetsLibraryGroupsEnumerationResultsBlock)enumerationBlock failureBlock:(ALAssetsLibraryAccessFailureBlock)failureBlock{
	BOOL ignore;
	enumerationBlock((ALAssetsGroup*) self, &ignore);
}

- (void)enumerateAssetsUsingBlock:(void(^)(ALAsset *asset, NSUInteger index, BOOL *stop))block{
	[self.mockAssets enumerateObjectsUsingBlock:block];
}

- (id)valueForProperty:(NSString *)property{
	return [NSString stringWithFormat:@"fake value for %@", property];
}

@end

@interface U1AssetManagerTest : CoreDataTestCase
@end

@implementation U1AssetManagerTest
{
	U1AssetManager *assetManager;
	dispatch_semaphore_t async_test_completed;
}

- (void)setUp{
	[super setUp];
	assetManager = [[U1AssetManager alloc] init];
	async_test_completed = dispatch_semaphore_create(0);
}

- (ALAsset *)mockAssetWithFileName:(NSString *)filename assetType:(NSString *)assetType{
	NSURL *url = [NSURL URLWithString:[NSString stringWithFormat:@"asset://%@", filename]];
	ALAssetRepresentation *fakeRepresentation = mock([ALAssetRepresentation class]);
	[given([fakeRepresentation valueForKey:@"filename"]) willReturn:filename];
	[given([fakeRepresentation filename]) willReturn:filename];
	[given([fakeRepresentation url]) willReturn:url];
	[given([fakeRepresentation valueForKey:@"url"]) willReturn:url];
	
	ALAsset *fakeAsset = mock([ALAsset class]);
	[given([fakeAsset defaultRepresentation]) willReturn:fakeRepresentation];
	[given([fakeAsset valueForProperty:ALAssetPropertyType]) willReturn:assetType];
	
	[given([fakeAsset defaultRepresentation]) willReturn:fakeRepresentation];
	[given([fakeAsset valueForKey:@"defaultRepresentation"]) willReturn:fakeRepresentation];
	return fakeAsset;
}

- (void)tearDown{
	[assetManager release];
	assetManager = nil;
	dispatch_release(async_test_completed);
	[super tearDown];
}

- (NSArray *)savedAssets{
	// grab saved Asset:
	NSFetchRequest *r = [NSFetchRequest fetchRequestWithEntityName:@"Asset"];
	NSSortDescriptor *sd = [NSSortDescriptor sortDescriptorWithKey:@"filename" ascending:YES];
	[r setSortDescriptors:@[sd]];
	NSError *error;
	NSArray *foundU1Assets = [ctx executeFetchRequest:r error:&error];
	if(foundU1Assets == nil){
		STFail(@"Error in fetch request: %@", error);
	}
	return foundU1Assets;
}

- (void)checkOneNewAssetWithFileName:(NSString *)filename{
	NSArray *foundU1Assets = [self savedAssets];
	if([foundU1Assets count] != 1){
		STFail(@"Should have one new asset, found %d", [foundU1Assets count]);
	}
	U1Asset *a = foundU1Assets[0];
	STAssertEqualObjects(a.filename, filename , @"");
}

#pragma mark Test Cases

- (void)testNoNewAssets{
	assetManager.assetsLibrary = (ALAssetsLibrary*)[[FakeAssetsLibrary alloc] initWithMockAssetArray:@[]];
	
	[assetManager updatePhotoAssetsWithAsyncBlock:^(U1Asset *asset, BOOL isNew){
		// do nothing in block, asset is saved by assetManager.
	}
    groupCompletionBlock:^(BOOL success){
		STAssertTrue(success, @"Error in updatePhotoAssetsWithAsyncBlock.");
		NSArray *foundU1Assets = [self savedAssets];
		if ([foundU1Assets count] != 0){
			STFail(@"should not have any assets, found %d", [foundU1Assets count]);
		}
		dispatch_semaphore_signal(async_test_completed);
	}];
	dispatch_semaphore_wait(async_test_completed, DISPATCH_TIME_FOREVER);
}

- (void)testOneNewAsset{
	ALAsset *fakeAsset = [self mockAssetWithFileName:@"fakephoto" assetType:ALAssetTypePhoto];
	assetManager.assetsLibrary = (ALAssetsLibrary*)[[FakeAssetsLibrary alloc] initWithMockAssetArray:@[fakeAsset]];
	
	[assetManager updatePhotoAssetsWithAsyncBlock:^(U1Asset *asset, BOOL isNew){
		// do nothing in block, asset is saved by assetManager.
	}
	groupCompletionBlock:^(BOOL success){
		STAssertTrue(success, @"Error in updatePhotoAssetsWithAsyncBlock.");
		[self checkOneNewAssetWithFileName:@"fakephoto"];
		dispatch_semaphore_signal(async_test_completed);
	}];
	dispatch_semaphore_wait(async_test_completed, DISPATCH_TIME_FOREVER);
}

- (void)testOneNewAssetPhotosOnly{
	ALAsset *fakeAsset = [self mockAssetWithFileName:@"fakephoto" assetType:ALAssetTypePhoto];
	ALAsset *fakeNonPhotoAsset = [self mockAssetWithFileName:@"fakevideo" assetType:ALAssetTypeVideo];
	
	assetManager.assetsLibrary = (ALAssetsLibrary*)[[FakeAssetsLibrary alloc] initWithMockAssetArray:@[fakeAsset, fakeNonPhotoAsset]];
	
	[assetManager updatePhotoAssetsWithAsyncBlock:^(U1Asset *asset, BOOL isNew){
	}
    groupCompletionBlock:^(BOOL success){
		STAssertTrue(success, @"Error in updatePhotoAssetsWithAsyncBlock.");

		[self checkOneNewAssetWithFileName:@"fakephoto"];
		dispatch_semaphore_signal(async_test_completed);
	}];
	dispatch_semaphore_wait(async_test_completed, DISPATCH_TIME_FOREVER);
}

- (void)testAddingAsset{
	ALAsset *fakeAsset = [self mockAssetWithFileName:@"fakephoto" assetType:ALAssetTypePhoto];
	assetManager.assetsLibrary = (ALAssetsLibrary*)[[FakeAssetsLibrary alloc] initWithMockAssetArray:@[fakeAsset]];
	
    [assetManager updatePhotoAssetsWithAsyncBlock:^(U1Asset *asset, BOOL isNew){
		// do nothing in block, asset is saved by assetManager.
	}
    groupCompletionBlock:^(BOOL success){
		STAssertTrue(success, @"Error in updatePhotoAssetsWithAsyncBlock.");
	
		[self checkOneNewAssetWithFileName:@"fakephoto"];
		dispatch_semaphore_signal(async_test_completed);
	}];
	dispatch_semaphore_wait(async_test_completed, DISPATCH_TIME_FOREVER);
	
	ALAsset *newAsset = [self mockAssetWithFileName:@"secondphoto" assetType:ALAssetTypePhoto];
	[(FakeAssetsLibrary*)assetManager.assetsLibrary addAsset:newAsset];
	
	[assetManager updatePhotoAssetsWithAsyncBlock:^(U1Asset *asset, BOOL isNew){
		// do nothing in block, asset is saved by assetManager.
	}
    groupCompletionBlock:^(BOOL success){
		STAssertTrue(success, @"Error in updatePhotoAssetsWithAsyncBlock.");
		
		
		NSArray *foundU1Assets = [self savedAssets];
		if([foundU1Assets count] != 2){
			STFail(@"Should have two assets, found %d", [foundU1Assets count]);
		}
		U1Asset *a = foundU1Assets[0];
		STAssertEqualObjects(a.filename, @"fakephoto" , @"");
		a = foundU1Assets[1];
		STAssertEqualObjects(a.filename, @"secondphoto" , @"");
		dispatch_semaphore_signal(async_test_completed);
	}];
	dispatch_semaphore_wait(async_test_completed, DISPATCH_TIME_FOREVER);
}


- (void)testCallTwiceNotAddingAsset{
	ALAsset *fakeAsset = [self mockAssetWithFileName:@"fakephoto" assetType:ALAssetTypePhoto];
	assetManager.assetsLibrary = (ALAssetsLibrary*)[[FakeAssetsLibrary alloc] initWithMockAssetArray:@[fakeAsset]];
	
	[assetManager updatePhotoAssetsWithAsyncBlock:^(U1Asset *asset, BOOL isNew){}
    groupCompletionBlock:^(BOOL success){
		STAssertTrue(success, @"Error in updatePhotoAssetsWithAsyncBlock.");
		[self checkOneNewAssetWithFileName:@"fakephoto"];
		dispatch_semaphore_signal(async_test_completed);
	}];
	dispatch_semaphore_wait(async_test_completed, DISPATCH_TIME_FOREVER);

	[assetManager updatePhotoAssetsWithAsyncBlock:^(U1Asset *asset, BOOL isNew){}
    groupCompletionBlock:^(BOOL success){
		STAssertTrue(success, @"Error in updatePhotoAssetsWithAsyncBlock.");
		[self checkOneNewAssetWithFileName:@"fakephoto"];
		dispatch_semaphore_signal(async_test_completed);
	}];
	dispatch_semaphore_wait(async_test_completed, DISPATCH_TIME_FOREVER);

}

@end
